from lcapy import *
import unittest
import sympy as sym


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_seriesRL(self):

        n1 = R(2) + L(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('seriesRL')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'R + L')

    def test_seriesRC(self):

        n1 = R(2) + C(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('seriesRC')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'R + C')

    def test_seriesGC(self):

        n1 = G(2) + C(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('seriesGC')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'G + C')

    def test_seriesLC(self):

        n1 = L(2) + C(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('seriesLC')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'L + C')

    def test_parallelRL(self):

        n1 = R(2) | L(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('parallelRL')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'R | L')

    def test_parallelRC(self):

        n1 = R(2) | C(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('parallelRC')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'R | C')

    def test_parallelGC(self):

        n1 = G(2) | C(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('parallelGC')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'G | C')

    def test_parallelLC(self):

        n1 = L(2) | C(3)
        Z1 = n1.Z(s)
        n2 = Z1.network('parallelLC')
        Z2 = n2.Z(s)

        self.assertEqual(Z1, Z2, 'L | C')

    def test_Z1(self):

        Z1 = LaplaceDomainImpedance(
            9 * s**5 + 30 * s**3 + 24 * s) / (18 * s**4 + 36 * s**2 + 8)

        self.assertEqual(Z1, Z1.network('fosterI').Z(s), 'fosterI')
        self.assertEqual(Z1, Z1.network('fosterII').Z(s), 'fosterII')
        self.assertEqual(Z1, Z1.network('cauerI').Z(s), 'cauerI')
        self.assertEqual(Z1, Z1.network('cauerII').Z(s), 'cauerII')

    def test_Z2(self):

        # Cauer II form.
        n = ((C(2) | L(3)) + C(4)) | L(5)
        Z1 = n.Z(s)

        self.assertEqual(Z1, Z1.network('fosterI').Z(s), 'fosterI')
        self.assertEqual(Z1, Z1.network('fosterII').Z(s), 'fosterII')
        self.assertEqual(Z1, Z1.network('cauerI').Z(s), 'cauerI')
        self.assertEqual(Z1, Z1.network('cauerII').Z(s), 'cauerII')

    def test_Z3(self):

        n = (R(2) | L(3)) + (R(4) | L(5))
        Z1 = n.Z(s)

        self.assertEqual(Z1, Z1.network('fosterI').Z(s), 'fosterI')
        self.assertEqual(Z1, Z1.network('fosterII').Z(s), 'fosterII')
        self.assertEqual(Z1, Z1.network('cauerI').Z(s), 'cauerI')
        self.assertEqual(Z1, Z1.network('cauerII').Z(s), 'cauerII')



"""This module provides the AngularFourierDomainExpression class to
represent omega-domain (angular frequency Fourier domain) expressions.

Copyright 2014--2022 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import AngularFourierDomain
from .expr import Expr, expr, expr_make
from .functions import integrate
from .inverse_fourier import inverse_fourier_transform
from .state import state, validate
from .sym import fsym, ssym, tsym, omegasym, omega0sym, j, pi, oo
from .units import u as uu
from sympy import Expr as symExpr

__all__ = ('omegaexpr', )


class AngularFourierDomainExpression(AngularFourierDomain, Expr):
    """Angular Fourier domain expression or symbol."""

    var = omegasym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)

        super(AngularFourierDomainExpression,
              self).__init__(val, **assumptions)

        expr = self.expr

        if check and not expr.has(omegasym):
            if expr.has(tsym):
                validate(state.t_in_w,
                         'omega-domain expression %s depends on t' % expr)
            if expr.has(ssym):
                validate(state.s_in_w,
                         'omega-domain expression % s depends on s' % expr)
            if expr.has(omegasym):
                validate(state.f_in_w,
                         'omega-domain expression %s depends on f' % expr)

    def as_expr(self):
        return AngularFourierDomainExpression(self)

    @property
    def energy(self):
        """Return signal energy."""

        E = integrate(abs(self)**2, (omega, -oo, oo)) / (2 * pi)

        # FIXME: result should not be an angular Fourier-domain expression
        return E

    def inverse_fourier(self, **assumptions):
        """Attempt inverse Fourier transform."""

        expr = self.subs(2 * pi * fsym)
        result = inverse_fourier_transform(expr.sympy, fsym, tsym)

        return self.change(result, 'time', units_scale=uu.Hz, **assumptions)

    def time(self, **assumptions):
        """Alias for inverse_fourier."""

        return self.inverse_fourier(**assumptions)

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""

        return self

    def fourier(self, **assumptions):
        """Convert to Fourier domain."""

        from .symbols import f

        result = self.subs(2 * pi * f)
        return result

    def norm_fourier(self, **assumptions):
        """Convert to normalized Fourier domain."""
        from .symbols import F
        from .sym import dt

        result = self.subs(2 * pi * F / dt)
        return result

    def norm_angular_fourier(self, **assumptions):
        """Convert to normalized angular Fourier domain."""
        from .symbols import Omega
        from .sym import dt

        result = self.subs(Omega / dt)
        return result

    def frequency_response(self, **assumptions):
        """Convert to frequency response domain."""

        return self.laplace(**assumptions).frequency_response()

    def angular_frequency_response(self, **assumptions):
        """Convert to angular frequency response domain."""

        return self.laplace(**assumptions).angular_frequency_response()

    def laplace(self, **assumptions):
        """Determine one-sided Laplace transform with 0- as the lower limit."""

        result = self.time(**assumptions).laplace()
        return result

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        return self.time(**assumptions).phasor(**assumptions)

    def plot(self, wvector=None, **kwargs):
        """Plot angular frequency response at values specified by wvector.

        There are many plotting options, see matplotlib.pyplot.plot.

        For example:
            V.plot(fvector, log_frequency=True)
            V.real.plot(fvector, color='black')
            V.phase.plot(fvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.

        `kwargs` include:
        `axes` - the plot axes to use otherwise a new figure is created
        `xlabel` - the x-axis label
        `ylabel` - the y-axis label
        `ylabel2` - the second y-axis label if needed, say for mag and phase
        `xscale` - the x-axis scaling, say for plotting as ms
        `yscale` - the y-axis scaling, say for plotting mV
        `norm` - use normalized frequency
        `dbmin` - the smallest value to plot in dB (default -120)
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned.  This is a tuple for magnitude/phase or
        real/imaginary plots.
        """

        from .plot import plot_angular_frequency
        return plot_angular_frequency(self, wvector, **kwargs)

    def bode_plot(self, wvector=None, unwrap=True, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations).  wvector
        specifies the angular frequencies.  If it is a tuple (f1, f2), it sets
        the frequency limits.   Since a logarithmic frequency scale is used,
        f1 must be greater than 0.

        `unwrap` controls phase unwrapping (default True).

        For more info, see `plot`.
        """

        from .plot import plot_angular_bode
        return plot_angular_bode(self, wvector, unwrap=unwrap, **kwargs)

    def nyquist_plot(self, wvector=None, log_frequency=True, **kwargs):
        """Plot frequency response as a Nyquist plot (imaginary part versus
        real part).  wvector specifies the angular frequencies.  If it
        is a tuple (f1, f2), it sets the frequency limits as (f1, f2).

        `npoints` set the number of plotted points.

        The unit circle is shown by default.  This can be disabled
        with `unitcircle=False`.

        """

        from .plot import plot_nyquist
        return plot_nyquist(self, wvector, log_frequency=log_frequency, **kwargs)

    def nichols_plot(self, wvector=None, log_frequency=True, **kwargs):
        """Plot frequency response as a Nichols plot (dB versus phase).
        wvector specifies the angular frequencies.  If it is a tuple
        (f1, f2), it sets the frequency limits as (f1, f2).

        `npoints` set the number of plotted points.

        """

        from .plot import plot_nichols
        return plot_nichols(self, wvector, log_frequency=log_frequency, **kwargs)


def omegaexpr(arg, **assumptions):
    """Create AngularFourierDomainExpression object.  If `arg` is omegasym return omega"""

    if arg is omegasym:
        return omega
    return expr_make('angular fourier', arg, **assumptions)


omega = AngularFourierDomainExpression('omega')
omega.units = uu.rad / uu.s



"""This module provides the Netlist class.  It could be rolled into
the Circuit class.

Copyright 2014--2023 Michael Hayes, UCECE

"""

# TODO: Add option to defer evaluation and thus keep things symbolic.
# This will help to simplify results that are not cancelled due to
# numerical quantisation.

from __future__ import division
from .admittance import admittance
from .config import solver_method
from .current import Iname, current
from .deprecation import LcapyDeprecationWarning
from .expr import Expr, expr, ExprList
from .mna import Nodedict, Branchdict
from .mnacpts import Cpt
from .netlistmixin import NetlistMixin
from .netlistopsmixin import NetlistOpsMixin
from .netlistsimplifymixin import NetlistSimplifyMixin
from .simulator import Simulator
from .subnetlist import SubNetlist
from .superpositionvoltage import SuperpositionVoltage
from .superpositioncurrent import SuperpositionCurrent
from .symbols import j, s, t, omega, omega0
from .transformdomains import TransformDomains
from .utils import isiterable
from copy import copy
from warnings import warn


class Netlist(NetlistOpsMixin, NetlistMixin, NetlistSimplifyMixin):
    """This class handles a generic netlist with multiple sources.
    During analysis, subnetlists are created for each source kind (dc,
    ac, transient, etc).  Since linearity is assumed, superposition is
    employed.

    """

    def __init__(self, filename=None, context=None, allow_anon=False):

        super(Netlist, self).__init__(filename, context, allow_anon=allow_anon)
        self._invalidate()
        self.kind = 'super'
        self.solver_method = solver_method

    @property
    def _time_kind(self):

        return self.kind in ('super', 'time', 't')

    def _groups(self):
        """Return dictionary of source groups keyed by domain.

        If the netlist is for an initial value problem, all the
        sources are in a single group called 'ivp'.  Any noise sources
        are ignored.

        If the netlist can be solved in the time-domain (i.e., there
        are no reactive components), all the sources are in single
        group called 'time'.

        Otherwise, the sources are decomposed and then grouped into
        the 'dc', 's', 'n*', and omega categories.  Note, a source can
        appear in multiple groups, for example, a source with voltage
        3 + u(t) will appear in both the 'dc' and 's' groups.

        """

        if self.is_IVP:

            def namelist(elements):
                return ', '.join([elt for elt in elements])

            if self.missing_ic != {}:
                warn('Missing initial conditions for %s' %
                     namelist(self.missing_ic))

            groups = self.independent_source_groups()
            newgroups = {'ivp': []}
            for key, sources in groups.items():
                if isinstance(key, str) and key[0] == 'n':
                    warn('Ignoring noise source %s for initial value problem' % sources)
                else:
                    newgroups['ivp'] += sources
            return newgroups

        elif self.is_time_domain:

            groups = self.independent_source_groups()
            newgroups = {'time': []}
            for key, sources in groups.items():
                if isinstance(key, str) and key[0] == 'n':
                    newgroups[key] = sources
                else:
                    newgroups['time'] += sources
            return newgroups

        else:
            return self.independent_source_groups(transform=True)

    def _subs_make(self, nowarn=False):

        if hasattr(self, '_sub'):
            return self._sub

        cct = self.expand()
        groups = cct._groups()
        sub = TransformDomains()

        for kind, sources in groups.items():
            sub[kind] = SubNetlist(cct, kind)

        if sub == {} and not nowarn:
            warn('Netlist has no sources')

        self._sub = sub
        return sub

    @property
    def cpts(self):
        """Return list of component names."""

        # Perhaps should prune wires, open-circuits, etc. ?
        return list(self._elements.keys())

    @property
    def kinds(self):
        """Return list of transform domain kinds required to analyse the netlist."""
        return list(self.sub.keys())

    @property
    def params(self):
        """Return list of symbols used as arguments in the netlist."""

        symbols = self.symbols
        params = []
        for elt in self.elements.values():
            for arg in elt.args:
                if arg in symbols and arg not in params:
                    params.append(arg)
        return params

    @property
    def sim(self):
        """Generate simulation object."""

        if hasattr(self, '_sim'):
            return self._sim

        self._sim = Simulator(self)
        return self._sim

    @property
    def ss(self):
        """Generate state-space representation.  See also `state_space()`"""
        return self.state_space()

    @property
    def sub(self):
        """Return dictionary of subnetlists keyed by transform domain kind.
        Note, the subnetlists are not created until a specific one is
        selected.

        """
        return self._subs_make()

    @property
    def subcircuits(self):
        """Return dictionary of subnetlists keyed by transform domain kind.
        Note, the subnetlists are not created until a specific one is
        selected.  The subcircuit keys are :
        'ivp' for an initial value problem solved using Laplace methods,
        's' for transient analysis using Laplace methods,
        'dc' for DC analysis,
        'time' for time-domain analysis when there are no reactive
        components,
        'n*' for noise-analysis (there is a subcircuit for
        each independent noise source), and
        omega (where omega is a number of expression specifying the angular
        frequency) for phasor analysis.

        """
        return self.sub

    @property
    def super_nodes(self):
        """Super nodes are nodes linked by voltage sources."""

        snodes = []

        for elt in self.elements.values():
            if not elt.is_voltage_source:
                continue
            snodes.append(elt.node_names[0:2])

        from .utils import merge_common

        # Merge supernodes to create super-supernodes...
        return list(merge_common(snodes))

    @property
    def symbols(self):
        """Return dictionary of symbols defined in the netlist."""

        return self.context.symbols

    @property
    def undefined_symbols(self):

        symbols = []
        for k, elt in self.elements.items():
            cpt = elt.cpt
            for arg in cpt.args:
                symbols += expr(arg).symbols

        return symbols

    @property
    def Idict(self):
        """Return dictionary of branch currents for each transform domain"""

        try:
            return self._Idict
        except AttributeError:
            pass

        result = Branchdict()
        for sub in self.sub.values():
            for node, value in sub.Idict.items():
                if node not in result:
                    result[node] = SuperpositionCurrent()
                result[node].add(value)
        self._Idict = result
        return result

    @property
    def Vdict(self):
        """Return dictionary of node voltages for each transform domain"""

        try:
            return self._Vdict
        except AttributeError:
            pass

        result = Nodedict()
        for sub in self.sub.values():
            for node, value in sub.Vdict.items():
                if node not in result:
                    result[node] = SuperpositionVoltage()
                result[node].add(value)

        self._Vdict = result
        return result

    def _add_ground(self, node=None):

        if '0' in self.nodes:
            return None

        if node is None:
            node = list(self.node_map)[0]
            warn('Ground node not specified: using node ' + node)

        return self.add('W %s 0' % node)

    def _add_test_voltage_source(self, Np, Nm):

        self._add('V? %s %s {DiracDelta(t)}' % (Np, Nm))
        return self.last_added()

    def _add_test_current_source(self, Np, Nm):

        self._add('I? %s %s {DiracDelta(t)}' % (Np, Nm))
        return self.last_added()

    def _cpt_add(self, cpt):

        if cpt.name in self._elements:
            warn('Overriding component %s' % cpt.name)
            # Need to search lists and update component.
            # For example, remove nodes that are only connected
            # to this component.
        else:
            # Check that this name won't conflict with an attr.
            # For example, cannot have name V or I.  Perhaps
            # rename these attributes?
            if hasattr(self, cpt.name):
                raise ValueError('Invalid component name %s' % cpt.name)

        self._elements[cpt.name] = cpt

        self._namespace_add(cpt.namespace)

    def get_I(self, name, nowarn=False):
        """Current through component (time-domain)"""

        self._add_ground()
        subs = self._subs_make(nowarn=nowarn)

        result = SuperpositionCurrent()
        for sub in subs.values():
            I = sub.get_I(name)
            result.add(I)
        result = result
        return result

    def get_i(self, name):
        """Time-domain current through component"""

        return self.get_I(name).time()

    def _get_Vd(self, Np, Nm=None, nowarn=False):
        """This does not check nodes."""

        self._add_ground()
        subs = self._subs_make(nowarn=nowarn)

        result = SuperpositionVoltage()
        for sub in subs.values():
            Vd = sub.get_Vd(Np, Nm)
            result.add(Vd)
        result = result.canonical()
        return result

    def get_Vd(self, Np, Nm=None, **kwargs):
        """Voltage drop between nodes (time-domain)"""

        Np, Nm = self._check_nodes(Np, Nm)
        return self._get_Vd(Np, Nm, **kwargs)

    def get_vd(self, Np, Nm=None):
        """Time-domain voltage drop between nodes"""

        return self.get_Vd(Np, Nm).time()

    def ac_omega_list(self):
        """Return list of the angular frequencies of the sources in
        the netlist."""

        omega_list = []
        for group in self.independent_source_groups(True).keys():
            if group in ('dc', 's'):
                continue
            if isinstance(group, str) and group[0] == 'n':
                continue
            omega_list.append(group)
        return omega_list

    def ac(self, omega=None):
        """Return netlist for ac components of independent sources
        for angular frequency `omega`.  If `omega` is undefined,
        the angular frequency `omega0` is used.

        See also: dc, transient, laplace.
        """

        if omega is None:

            omega_list = self.ac_omega_list()
            if len(omega_list) > 1:
                omega = omega_list[0]
                warn('Netlist has multiple AC frequencies: %s, using %s' %
                     (omega_list, omega))
            elif len(omega_list) == 1:
                omega = omega_list[0]
            else:
                # Dummy, it could be anything.
                omega = 1

        return self.select(omega)

    def annotate_currents(self, cpts=None, domainvar=None, flow=False,
                          eng_format=True, evalf=True, num_digits=3,
                          show_units=True, pos=''):
        """Annotate specified list of component names `cpts` with current (or
        flow).

        `domainvar` specifies the domain to calculate the voltages for
        (e.g., `t` for time-domain, `s` for Laplace-domain)

        `flow` (default False) if True annotates current as a flow

        `eng_format` (default True) if True use engineering format if
        the current is a number, e.g., 100\, mV instead of 0.1\, V

        `evalf` (default True) if True prints floating point
        numbers as decimals otherwise they are shown as rationals

        `show_units` (default True) if True applies the units(e.g.,
        V for volts)

        `pos` specifies where to position the labels(see docs)
        """

        cct = self.copy()
        groundwire = cct._add_ground()
        if cpts is None:
            cpts = []
            for elt in cct._elements.values():
                if (elt.is_resistor or elt.is_capacitor or
                        elt.is_inductor or elt.is_voltage_source):
                    cpts.append(elt.name)

        label = ('f' if flow else 'i') + pos

        if domainvar is None:
            domainvar = t

        new = self._new()
        for cpt in cct._elements.values():
            net = cpt._copy()
            if cpt.name in cpts:
                I = cpt.I(domainvar)
                if evalf:
                    I = I.evalf(num_digits)
                net += ', ' if ';' in net else '; '
                net += ', %s={$%s$}' % (label, I.latex_with_units(
                    eng_format=eng_format, evalf=evalf, num_digits=num_digits, show_units=show_units))
            new.add(net)
        if groundwire is not None:
            new.remove(groundwire.name)
        return new

    def annotate_voltages(self, cpts=None, domainvar=None,
                          eng_format=True, evalf=True, num_digits=3,
                          show_units=True, pos=''):
        """Annotate specified list of component names `cpts` with voltage.

        `domainvar` specifies the domain to calculate the voltages for
        (e.g., `t` for time-domain, `s` for Laplace-domain)

        `pos` specifies where to position the labels, see docs

        `eng_format` (default True) if True use engineering format if
        the voltage is a number, e.g., 100\, mV instead of 0.1\, V

        `evalf` (default True) if True prints floating point
        numbers as decimals otherwise they are shown as rationals

        `show_units` (default True) if True applies the units(e.g.,
        V for volts)

        `pos` specifies where to position the labels(see docs)
        """

        cct = self.copy()
        groundwire = cct._add_ground()
        if cpts is None:
            cpts = []
            for elt in cct._elements.values():
                if (elt.is_resistor or elt.is_capacitor or
                        elt.is_inductor or elt.is_current_source):
                    cpts.append(elt.name)

        if domainvar is None:
            domainvar = t

        new = self._new()
        for cpt in cct._elements.values():
            net = cpt._copy()
            if cpt.name in cpts:
                V = cpt.V(domainvar)
                if evalf:
                    V = V.evalf(num_digits)
                net += ', ' if ';' in net else '; '
                net += 'v%s={$%s$}' % (pos, V.latex_with_units(eng_format=eng_format,
                                       evalf=evalf, num_digits=num_digits, show_units=show_units))
            new.add(net)
        if groundwire is not None:
            new.remove(groundwire.name)
        return new

    def annotate_node_voltages(self, nodes=None, domainvar=None,
                               label_voltages=False, eng_format=True,
                               evalf=True, num_digits=3,
                               show_units=True, anchor='south west'):
        """Create a new netlist with the node voltages annotated.  This is
        useful for drawing a schematic with the node voltages shown.
        For example,

        `cct.annotate_node_voltages((1, 2, 3)).draw()`

        `nodes` is a list of the nodes to annotate or `None` for all.

        `domainvar` specifies the domain to calculate the voltages for
        (e.g., `t` for time-domain, `s` for Laplace-domain)

        `label_voltages` (default False) if True prefixes the
        annotation with V1 = for node 1, etc.

        `eng_format` (default True) if True use engineering format if
        the voltage is a number, e.g., 100\, mV instead of 0.1\, V

        `evalf` (default True) if True prints floating point
        numbers as decimals otherwise they are shown as rationals

        `num_digits` (default 3) specfies the number of digits to print
        for floating point numbers

        `show_units` (default True) if True applies the units(e.g.,
        V for volts)

        `anchor` (default 'south west') specifies the position of the
        voltage label

        """

        cct = self.copy()
        groundwire = cct._add_ground()
        if nodes is None:
            nodes = cct.node_list
        elif not isiterable(nodes):
            nodes = (nodes, )

        if domainvar is None:
            domainvar = t

        new = self.copy()
        for node in nodes:
            v = cct[node].V(domainvar)
            if evalf:
                v = v.evalf(num_digits)

            vstr = '%s' % v.latex_with_units(eng_format=eng_format, evalf=evalf,
                                             num_digits=num_digits,
                                             show_units=show_units)

            if label_voltages:
                vstr = 'V_{%s}=' % node + vstr

            new.add('A%s %s; l={%s}, anchor=%s' % (node, node, vstr, anchor))
        if groundwire is not None:
            new.remove(groundwire.name)
        return new

    def annotate_voltage(self, cpts, domainvar=None, pos=''):
        LcapyDeprecationWarning(
            feature="annotate_voltage",
            useinstead="annotate_voltages",
            deprecated_since_version="1.0",
            issue=61
        ).warn()
        return self.annotate_voltages(cpts, domainvar=domainvar,
                                      pos=pos, evalf=True, num_digits=3)

    def annotate_current(self, cpts, domainvar=None, flow=False, pos=''):
        LcapyDeprecationWarning(
            feature="annotate_current",
            useinstead="annotate_currents",
            deprecated_since_version="1.0",
            issue=61
        ).warn()
        return self.annotate_currents(cpts, domainvar=domainvar, flow=flow,
                                      pos=pos, evalf=True, num_digits=3)

    def apply_test_current_source(self, Np, Nm=None):
        """This copies the netlist, kills all the sources, and applies a Dirac
        delta test current source across the specified nodes.  If the
        netlist is not connected to ground, the negative specified
        node is connected to ground.  The new netlist is returned."""

        Np, Nm = self._parse_node_args2(Np, Nm)
        Np, Nm = self._check_nodes(Np, Nm)

        new = self.kill()
        new._add_ground(Nm)
        new._add_test_current_source(Np, Nm)
        return new

    def apply_test_voltage_source(self, Np, Nm=None):
        """This copies the netlist, kills all the sources, and applies a Dirac
        delta test voltage source across the specified nodes.  If the
        netlist is not connected to ground, the negative specified
        node is connected to ground.  The new netlist is returned."""

        Np, Nm = self._parse_node_args2(Np, Nm)
        Np, Nm = self._check_nodes(Np, Nm)

        new = self.kill()
        new._add_ground(Nm)
        new._add_test_voltage_source(Np, Nm)
        return new

    def dc(self):
        """Return netlist for dc components of independent sources.

        See also: ac, transient, laplace.
        """
        return self.select('dc')

    def describe(self):
        """Print a message describing how circuit is solved."""
        print(self.description())

    def description(self):
        """Return a message describing how circuit is solved."""

        def describe_sources(sources, omega=None):
            sources_string = ', '.join(sources)
            if len(sources) == 1:
                s = 'source %s' % sources_string
            else:
                s = 'sources %s' % sources_string

            if omega is not None:
                s += ' at angular frequency ' + str(omega)
            return s

        def describe_analysis(method, sources, omega=None):
            return '%s analysis is used for %s.\n' % (method,
                                                      describe_sources(sources, omega))

        if self.is_switching:
            return 'This has switches and thus is time variant.  Use the convert_IVP(t) method to convert to an initial value problem, specifying the time when to evaluate the switches.'

        groups = self.independent_source_groups(
            transform=not self.is_time_domain)

        if groups == {}:
            return 'There are no non-zero independent sources so everything is zero.\n'

        if self.is_IVP:
            return 'This has initial conditions for %s so is an initial value problem solved in the Laplace-domain using Laplace transforms.\n' \
                % ', '.join(self.ics)
            return

        s = ''
        if len(groups) > 1:
            s = 'This is solved using superposition.\n'

        for kind, sources in groups.items():
            if not isinstance(kind, str):
                s += describe_analysis('Phasor', sources, kind)
            elif kind[0] == 'n':
                s += describe_analysis('Noise', sources)
            elif kind == 'dc':
                s += describe_analysis('DC', sources)
            elif kind == 's':
                s += describe_analysis('Laplace', sources)
            elif kind in ('t', 'time'):
                s += describe_analysis('Time-domain', sources)
        return s

    def expand(self):
        """Expand the netlist, replacing complicated components with simpler
        components."""

        new = self._new()

        for cpt in self._elements.values():
            new._add(cpt._expand())
        return new

    def laplace(self):
        """Return netlist for Laplace representations of independent
        source values.

        See also: dc, ac, transient.

        """
        return self.select('laplace')

    def transient(self):
        """Return netlist for transient components of independent
        sources.  Note, unlike the similar laplace method, dc and ac
        components are ignored.

        See also: dc, ac, laplace.

        """
        return self.select('s')

    def _new(self):

        # TODO.  Copy or share?
        context = self.context
        return self.__class__(context=context)

    def remove(self, name):
        """Remove specified element by name or elements specified in list."""

        if name is None:
            return self

        if isinstance(name, (list, tuple)):
            for name1 in name:
                self.remove(name1)
            return self

        if name not in self._elements:
            raise ValueError('Unknown component: ' + name)

        self._invalidate()

        cpt = self._elements[name]
        for node in cpt.nodes:
            node.remove(cpt)

        self._elements.pop(name, None)
        return self

    def annotate(self, cpts, *args, **kwargs):
        """Annotate a particular component (or list of components)
        with specified schematic attributes and return new netlist.

        For example:
        `cct.annotate('R1', color='blue')`
        `cct.annotate('R1', 'color=blue, dashed')`
        `cct.annotate(('U1', 'U2'), fill='blue')`

        See also `highlight`."""

        if not isinstance(cpts, (tuple, list, set)):
            cpts = [cpts]

        names = []
        for cpt in cpts:
            if isinstance(cpt, Cpt):
                name = cpt.name
            else:
                name = cpt
            if not self.has(name):
                raise ValueError('Unknown component %s' % name)
            names.append(name)

        new = self._new()

        for cpt in self._elements.values():
            if cpt.name in names:
                new.add(cpt.annotate(*args, **kwargs))
            else:
                new.add(cpt._copy())
        return new

    def highlight(self, cpts, color='blue'):
        """Highlight a particular component (or list of components)
        with specified color and return new netlist.

        See also `annotate`."""

        return self.annotate(cpts, color=color)

    def matrix_equations(self, form='default', invert=False):
        """Return modified nodal analysis equations in matrix form.

        Forms can be:
         'default'
         'A y = b'
         'b = A y'
         'Ainv b = y'
         'y = Ainv b'

        If `invert` is True, evaluate the matrix inverse."""

        mna = self.modified_nodal_analysis()
        return mna.matrix_equations(form, invert)

    def modified_nodal_analysis(self):
        """Perform modified nodal analysis for this netlist.  This is
        cached."""

        if hasattr(self, '_mna'):
            return self._mna

        if self.kind in ('time', 'super'):
            raise ValueError(
                'Cannot put time domain equations into matrix form.  '
                'Convert to dc, ac, or laplace domain first.')

        self._mna = SubNetlist(self, self.kind).mna

        return self._mna

    def nodal_analysis(self):
        """Perform nodal analysis for this netlist.   This is cached."""

        from .nodalanalysis import NodalAnalysis

        if hasattr(self, '_na'):
            return self._na

        self._na = NodalAnalysis.from_circuit(self)
        return self._na

    def noisy_except(self, *args, T='T'):
        """Return a new circuit with all but the specified resistors in series
        with noise voltage sources"""

        for arg in args:
            if arg not in self.elements and not self.elements[arg].is_resistor:
                raise ValueError('Element %s is not a known resistor' % arg)
        resistors = []
        for cpt in self.elements.values():
            if cpt.is_resistor and not cpt.is_noiseless and cpt.name not in args:
                resistors.append(cpt.name)
        return self._noisy(resistors, T)

    def noise_model(self, T='T'):
        """"Create noise model where resistors are converted into a series
        combination of an ideal resistor and a noise voltage
        source."""

        return self.noisy(T=T)

    def noisy(self, *args, T='T'):
        """Return a new circuit with the specified resistors in series
        with noise voltage sources"""

        if len(args) == 0:
            return self.noisy_except(T=T)

        resistors = []
        for arg in args:
            if arg not in self.elements and not self.elements[arg].is_resistor:
                raise ValueError('Element %s is not a known resistor' % arg)
            resistors.append(arg)

        return self._noisy(resistors, T=T)

    def initialize(self, cct, T=None):
        """Set the initial values for this netlist based on the values
        computed for netlist `cct` at specified time `T`.

        Alternatively, set the initial values using a dictionary
        of values keyed by the component name.
        """

        if isinstance(cct, dict):
            return self._initialize_from_dict(cct)

        return self._initialize_from_circuit(cct, T)

    def loop_analysis(self):
        """Perform loop analysis for this netlist.   This is cached.

        This is currently an alias for `mesh_analysis()` and so only works
        for circuits with a planar topology.
        """

        return self.mesh_analysis()

    def mesh_analysis(self):
        """Perform mesh analysis for this netlist.   This is cached.

        This is only applicable for circuits with a planar topology.
        """

        from .loopanalysis import LoopAnalysis

        if hasattr(self, '_la'):
            return self._la

        self._la = LoopAnalysis.from_circuit(self)
        return self._la

    def save(self, filename):
        """Save netlist to file."""

        f = open(filename, 'w')
        f.writelines(str(self))
        f.close()

    def select(self, kind):
        """Return new netlist with transform domain kind selected for
        specified sources in sourcenames.

        """

        new = self._new()
        new.kind = kind

        for cpt in self._elements.values():
            net = cpt._select(kind)
            new._add(net)
        return new

    def open_circuit(self, cpt):
        """Apply open-circuit in series with component.  Returns name of open
        circuit component."""

        if isinstance(cpt, Cpt):
            cpt = cpt.name
        return self.elements[cpt].open_circuit()

    def short_circuit(self, cpt):
        """Apply short-circuit across component.  Returns name of voltage
        source component used as the short."""

        if isinstance(cpt, Cpt):
            cpt = cpt.name
        return self.elements[cpt].short_circuit()

    def convert_IVP(self, t=0):
        """Remove switches from netlist and convert to an initial value
        problem. `t` is used to determine the state of the switches."""

        times = self.switching_times()
        if times == ():
            warn('Netlist has no switches')
            return self

        if t < times[0]:
            return self.replace_switches(t)

        cct = self
        for m, time in enumerate(times):
            if time > t:
                break
            before = cct.replace_switches_before(time)
            cct = cct.replace_switches(time).initialize(before, time)

        if time != 0:
            warn('Note, the time t is relative to %s' % time)

        return cct

    def renumber(self, node_map=None):
        """Renumber nodes using specified node_map.  If node_map not specified
        then a mapping is created."""

        if node_map is None:
            node_map = {}

        if len(node_map) != len(self.nodes):
            node_map = self.augment_node_map(node_map)

        new = self._new()

        for cpt in self._elements.values():
            new._add(cpt._rename_nodes(node_map))
        return new

    def replace(self, oldname, newname):
        """Replace component.

        For example,
        b = a.replace('C', 'W')
        c = a.replace('C1', 'C1 1 2')
        """

        new = self._new()

        newparts = newname.split(' ')

        for cpt in self._elements.values():
            net = cpt._copy()
            if cpt.name == oldname:
                if len(newparts) == 1:
                    # Just replace name of component
                    parts = net.split(' ')
                    parts[0] = newname
                    net = ' '.join(parts)
                else:
                    # Replace with new net
                    net = newname
            new._add(net)
        return new

    def _replace_switches(self, t=0, switchnames=None, before=False):
        """Replace specified switches with wire or open circuit
        for time `t`.   If `switchnames` is not specified, all switches
        are replaced."""

        new = self._new()

        for cpt in self._elements.values():
            if switchnames is None or cpt.name in switchnames:
                net = cpt._replace_switch(t, before=before)
            else:
                net = cpt._copy()
            new._add(net)
        return new

    def replace_switches(self, t=0, switchnames=None):
        """Replace specified switches with open-circuit or short-circuit for
        time at or after `t`.  If `switchnames` is not specified, all
        switches are replaced."""

        return self._replace_switches(t, switchnames, before=False)

    def replace_switches_before(self, t=0, switchnames=None):
        """Replace specified switches with open-circuit or short-circuit for
        time just before `t`.  If `switchnames` is not specified, all
        switches are replaced."""

        return self._replace_switches(t, switchnames, before=True)

    def switching_times(self, tmax=1e12):
        """Return sorted list of the times that switches activate prior to
        `tmax`."""

        times = []
        for cpt in self._elements.values():
            if cpt.type.startswith('SW'):
                active_time = float(cpt.args[0])
                if active_time < tmax and active_time not in times:
                    times.append(active_time)
        return sorted(times)

    def branch_current_names(self):
        """Return `ExprList` of branch current names of the form i_cptname
        for the time-domain and of the form I_cptname othwerwise."""

        branch_list = self.branch_list

        prefix = 'I'
        if self._time_kind:
            prefix = 'i'

        current_name_list = ExprList()
        for branch in branch_list:
            current_name_list.append(prefix + branch)

        return current_name_list

    def branch_currents(self):
        """Return `ExprList` of branch currents.  Each element is a
        SuperpositionVoltage object.

        If you want a vector of time-domain expressions use
        `Vector(cct.branch_currents()(t))`.
        """

        return ExprList([self[b].I for b in self.branch_list])

    def branch_voltage_names(self):
        """Return `ExprList` of branch voltage names of the form v_cptname
        for the time-domain and of the form V_cptname othwerwise."""

        branch_list = self.branch_list

        prefix = 'V'
        if self._time_kind:
            prefix = 'v'

        voltage_name_list = ExprList()
        for branch in branch_list:
            voltage_name_list.append(prefix + branch)

        return voltage_name_list

    def branch_voltages(self):
        """Return `ExprList` of branch voltages.  Each element is a
        SuperpositionVoltage object.

        If you want a vector of time-domain expressions use
        `Vector(cct.branch_voltages()(t))`.
        """

        return ExprList([self[b].V for b in self.branch_list])

    def evidence_matrix(self):
        """Return the evidence matrix.  This has a size NxM where
        N is the number of nodes and M is the number of branches.

        An element is 1 for currents entering a node from a branch, -1
        for currents leaving a node from a branch, and zero otherwise.

        Note, each column sums to zero.

        """

        from .matrix import Matrix
        from sympy import zeros

        cct = self

        branch_list = cct.branch_list
        current_list = self.branch_currents

        nodes = list(cct.equipotential_nodes)
        node_map = cct.node_map

        mat = Matrix(zeros(len(nodes), len(branch_list)))

        for m, cpt_name in enumerate(branch_list):
            cpt = cct[cpt_name]
            node1, node2 = cpt.nodes[0:2]
            n1 = nodes.index(node_map[node1.name])
            n2 = nodes.index(node_map[node2.name])

            mat[n1, m] = 1
            mat[n2, m] = -1

        return mat



"""
This module implements modified nodal analysis (MNA).

Copyright 2014--2023 Michael Hayes, UCECE
"""

from __future__ import division
from .assumptions import Assumptions
from .vector import Vector
from .matrix import Matrix, matrix_solve
from .sym import eps
from .expr import ExprDict, expr
from .voltage import Vtype
from .current import Itype, current_sign
from .systemequations import SystemEquations
import sympy as sym
from warnings import warn

# Note, all the maths is performed using sympy expressions and the
# values and converted to Expr when required.  This is more
# efficient and, more importantly, overcomes some of the wrapping
# problems which casues the is_real attribute to be dropped.


class Nodedict(ExprDict):

    def __getitem__(self, name):
        """Return node by name or number."""

        # If name is an integer, convert to a string.
        if isinstance(name, int):
            name = '%d' % name
        return super(Nodedict, self).__getitem__(name)


class Branchdict(ExprDict):
    pass


class MNA(object):
    """This class performs modified nodal analysis (MNA) on a netlist of
    components.  There are several variants:

    1. DC analysis if all the independent sources are DC.  The .V and .I
    methods return DC expressions with the dc assumption set.

    2. AC analysis if all the independent sources are AC.  The .V and .I
    methods return phasors.

    3. Initial value Laplace analysis if an L or C has an explicit
    initial value.  The .V and .I methods return s-domain expressions
    with no assumption sets; thus the time-domain results are only
    valid for t >= 0.

    4. General Laplace analysis.  If all the sources are causal and
    all the initial conditions are zero (explicitly or implicitly)
    then the time-domain results are causal.

    5. Noise analysis.

    Note, it is assumed that the user of this class uses superposition
    to solve problems with mixed independent sources, such as DC and
    AC.

    """

    def __init__(self, cct, solver_method):

        self.cct = cct
        self.kind = cct.kind
        self.solver_method = solver_method

        if cct.elements == {}:
            raise ValueError('No elements to analyse')

        # TODO: think this out.  When a circuit is converted
        # to a s-domain model we get Z (and perhaps Y) components.
        # We also lose the ability to determine the voltage
        # across a capacitor or inductor since they get split
        # into a Thevenin model and renamed.
        if hasattr(self, '_s_model'):
            raise RuntimeError('Cannot analyse s-domain model')

        # Determine which branch currents are needed.
        self.unknown_branch_currents = []

        for elt in self.cct.elements.values():
            if elt.need_branch_current:
                self.unknown_branch_currents.append(elt.name)
            if elt.need_extra_branch_current:
                self.unknown_branch_currents.append(elt.name + 'X')
            if elt.is_current_controlled:
                cname = elt.args[0]
                if cname not in self.cct.elements:
                    raise ValueError(
                        'Undefined controlling source %s for %s' % (cname, elt.name))
                if cname not in self.unknown_branch_currents:
                    self.unknown_branch_currents.append(cname)

        # Generate stamps.
        num_nodes = len(self.cct.node_list) - 1
        num_branches = len(self.unknown_branch_currents)

        self._G = sym.zeros(num_nodes, num_nodes)
        self._B = sym.zeros(num_nodes, num_branches)
        self._C = sym.zeros(num_branches, num_nodes)
        self._D = sym.zeros(num_branches, num_branches)

        self._Is = sym.zeros(num_nodes, 1)
        self._Es = sym.zeros(num_branches, 1)

        # Iterate over circuit elements and fill in matrices.
        for elt in self.cct.elements.values():
            if not elt.nosim:
                elt._stamp(self)

        # Augment the admittance matrix to form A matrix.
        self._A = self._G.row_join(self._B).col_join(self._C.row_join(self._D))
        # Augment the known current vector with known voltage vector
        # to form Z vector.
        self._Z = self._Is.col_join(self._Es)

    def _invalidate(self):
        for attr in ('_A', '_Vdict', '_Idict'):
            if hasattr(self, attr):
                delattr(self, attr)

    def _cpt_node_indexes(self, cpt):

        return [self._node_index(n) for n in cpt.node_names]

    def _cpt_branch_index(self, cpt):

        return self._branch_index(cpt.name)

    def _node_index(self, node):
        """Return node index; ground is -1"""
        return self.cct.node_list.index(self.cct.node_map[node]) - 1

    def _branch_index(self, cpt_name):

        try:
            index = self.unknown_branch_currents.index(cpt_name)
            return index
        except ValueError:
            raise ValueError(
                'Unknown component name %s for branch current' % cpt_name)

    def _failure_reasons(self):

        message = 'The MNA A matrix is not invertible for %s analysis:' % self.kind
        cct = self.cct
        if cct.nodes['0'].count == 0:
            return message + ' Nothing electrically connected to ground.'

        if not cct.is_connected:
            return message + ' Not all nodes are connected.  Use cct.unconnected_nodes() to find them.'

        reasons = []
        components = cct.components
        if cct.kind == 'dc':
            reasons.append('Check there is a DC path between all nodes.')
        if components.transformers != []:
            reasons.append(
                'Check secondary of transformer is referenced to ground.')
        if len(components.capacitors) > 1:
            reasons.append('Check capacitors are not in series.')
        if components.voltage_sources != []:
            reasons.append('Check voltage source is not short-circuited.')
        if len(components.voltage_sources) > 1:
            reasons.append('Check for loop of voltage sources.')
        if components.current_sources != []:
            reasons.append('Check current source is not open-circuited.')
        if len(components.current_sources) > 1:
            reasons.append('Check for current sources in series.')

        return message + '\n    ' + '\n    '.join(reasons)

    def _solve(self):
        """Solve network.  This does not simplify the results."""

        if hasattr(self, '_Vdict'):
            return

        cct = self.cct

        if '0' not in cct.node_map:
            raise ValueError('No ground node specified.')

        solver_method = self.solver_method
        # Hack to overcome simplification of sqrt with noise circuits
        # if using LU solver. For example,
        # opamp-noninverting-amplifier-noise1.py
        if (solver_method in ('LU', 'GE') and isinstance(self.kind, str)
                and self.kind[0] == 'n'):
            # GE is faster but produces sqrt(x**2 + 2*x*y + y**2)
            # that SymPy does not realise is x + y.  This causes
            # a failure with test_noisy2.
            # LU suffers from similar problems and can make a dog's breakfast
            # ADJ seems robust but is slower than GE.
            solver_method = 'ADJ'

        # Solve for the nodal voltages
        try:
            results = matrix_solve(self._A, self._Z, method=solver_method)
        except ValueError:
            message = self._failure_reasons()
            raise ValueError(message)

        results = results.subs(cct.context.symbols)

        # Handle capacitors at DC by assuming an infinite resistance
        # in parallel.
        if results.has(eps):
            # Perhaps this should be lazily done if needed
            results = results.limit(eps, 0)

        branchdict = {}
        for elt in cct.elements.values():
            if elt.type in ('K', 'Cable') or elt.ignore:
                continue
            n1 = cct.node_map[elt.node_names[0]]
            n2 = cct.node_map[elt.node_names[1]]
            branchdict[elt.name] = (n1, n2)

        vtype = Vtype(self.kind)
        itype = Itype(self.kind)
        assumptions = Assumptions()
        if vtype.is_phasor_domain:
            assumptions.set('omega', self.kind)
        elif self.kind in ('s', 'ivp'):
            assumptions.set('ac', cct.is_ac)
            assumptions.set('dc', cct.is_dc)
            assumptions.set('causal', cct.is_causal)
        elif isinstance(self.kind, str) and self.kind[0] == 'n':
            assumptions.set('nid', self.kind)

        # Create dictionary of node voltages
        self._Vdict = Nodedict()
        self._Vdict['0'] = vtype(0, **assumptions)
        for n in cct.nodes:
            index = self._node_index(n)
            if index >= 0:
                self._Vdict[n] = vtype(
                    results[index], **assumptions)
            else:
                self._Vdict[n] = vtype(0, **assumptions)

        num_nodes = len(cct.node_list) - 1

        # Create dictionary of branch currents through elements
        self._Idict = Branchdict()
        for m, key in enumerate(self.unknown_branch_currents):
            I = current_sign(
                results[m + num_nodes], cct.elements[key].is_source)
            self._Idict[key] = itype(I, **assumptions)

        # Calculate the branch currents.  These should be lazily
        # evaluated as required.
        for elt in cct.elements.values():
            if elt.type in ('R', 'NR', 'C'):
                n1 = cct.node_map[elt.node_names[0]]
                n2 = cct.node_map[elt.node_names[1]]
                V1, V2 = self._Vdict[n1], self._Vdict[n2]
                I = (V1.expr - V2.expr - elt.V0.expr) / elt.Z.expr
                self._Idict[elt.name] = itype(I, **assumptions)
            elif elt.type in ('I', ):
                I = current_sign(-elt.Isc, True)
                self._Idict[elt.name] = I

    @property
    def A(self):
        """Return A matrix for MNA"""

        ret = Matrix(self._A)

        # Could cache
        if ret.has(eps):
            # Should only do this after inverse...  It is more rigorous
            # if don't evaluate limit.
            ret = ret.limit(eps, 0)
        return ret

    @property
    def B(self):
        """Return B matrix for MNA"""

        return Matrix(self._B)

    @property
    def C(self):
        """Return C matrix for MNA"""

        return Matrix(self._C)

    @property
    def D(self):
        """Return D matrix for MNA"""

        return Matrix(self._D)

    @property
    def G(self):
        """Return G matrix for MNA"""

        return Matrix(self._G)

    @property
    def Z(self):
        """Return Z vector for MNA"""

        return Vector(self._Z)

    @property
    def E(self):
        """Return E vector for MNA"""

        return Vector(self._Es)

    @property
    def I(self):
        """Return I vector for MNA"""

        return Vector(self._Is)

    @property
    def X(self):
        """Return X vector (of unknowns) for MNA"""

        V = [self.cct.Vname('Vn%s' % node) for node in self.cct.node_list[1:]]
        I = [self.cct.Iname('I%s' % branch)
             for branch in self.unknown_branch_currents]
        return Vector(V + I)

    @property
    def Vdict(self):
        """Return dictionary of transform domain node voltages indexed by node
        name"""

        self._solve()
        return self._Vdict

    @property
    def Idict(self):
        """Return dictionary of transform domain branch currents indexed by
        component name"""

        self._solve()
        return self._Idict

    def matrix_equations(self, form='default', invert=False):
        """System of equations used to find the unknowns.

        Forms can be:
         A y = b
         b = A y
         Ainv b = y
         y = Ainv b

        If `invert` is True, evaluate the matrix inverse."""

        sys = SystemEquations(self.A, self._Z, self.X)
        return sys.format(form, invert)

    def equations(self, inverse=False):
        """System of equations used to find the unknowns.

        If inverse is True, evaluate the matrix inverse.

        This is for compatibility and is deprecated.  Use
        matrix_equations instead."""

        return self.matrix_equations(invert=inverse)



"""This module provides support for the unilateral inverse Laplace
transform.

Copyright 2021--2022 Michael Hayes, UCECE

"""

from .transformer import UnilateralInverseTransformer
from .ratfun import Ratfun
from .sym import simplify, AppliedUndef
from .utils import factor_const, scale_shift, as_sum_terms
from .matrix import matrix_inverse
from warnings import warn
import sympy as sym

Zero = sym.S.Zero
One = sym.S.One

__all__ = ('ILT', 'inverse_laplace_transform')


class InverseLaplaceTransformer(UnilateralInverseTransformer):

    name = 'inverse Laplace transform'

    def noevaluate(self, expr, s, t):

        # Construct Bromwich integral.
        self.error('TODO')

    def check(self, expr, s, t, **kwargs):

        if expr.has(t):
            for term in expr.as_ordered_terms():
                # Allow Subs(Derivativee(undef, t), t, 0)
                if term.has(t) and not term.is_constant:
                    self.error('Expression depends on t')

    def key(self, expr, s, t, **kwargs):
        return (expr, s, t,
                kwargs.get('causal', False),
                kwargs.get('zero_initial_conditions', True),
                kwargs.get('damped_sin', True),
                kwargs.get('damping', None))

    def func(self, expr, s, t):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], s)

        # Convert V(s) to v(t), etc.
        name = expr.func.__name__
        undef = sym.Function(name[0].lower() + name[1:])(t)
        result = undef.subs(t, t / scale) / abs(scale)

        if shift != 0:
            result = result * sym.exp(t * shift / scale)
        return result

    def do_damped_sin(self, expr, s, t):

        ncoeffs, dcoeffs = expr.coeffs()
        K = ncoeffs[0] / dcoeffs[0]

        ncoeffs = [(c / ncoeffs[0]) for c in ncoeffs]
        dcoeffs = [(c / dcoeffs[0]) for c in dcoeffs]

        if len(ncoeffs) > 3 or len(dcoeffs) > 3:
            self.error('Not a second-order response')

        omega0 = sym.sqrt(dcoeffs[2])
        zeta = dcoeffs[1] / (2 * omega0)

        if zeta.is_constant() and zeta > 1:
            warn('Expression is overdamped')

        sigma1 = (zeta * omega0).simplify()
        omega1 = (omega0 * sym.sqrt(1 - zeta**2)).simplify()
        K = (K / omega1).simplify()

        E = sym.exp(-sigma1 * t)
        S = sym.sin(omega1 * t)

        h = K * E * S

        # If overdamped
        #h = K * sym.exp(-sigma1 * t) * sym.sinh(omega0 * mu * t)

        if len(ncoeffs) == 1:
            return Zero, h

        C = sym.cos(omega1 * t)
        kCd = omega1
        kSd = -sigma1
        hd = K * E * (kCd * C + kSd * S)

        if len(ncoeffs) == 2:
            return Zero, K * E * (kCd * C + (ncoeffs[1] + kSd) * S)

        kCdd = -2 * omega1 * sigma1
        kSdd = sigma1**2 - omega1**2

        G = K * E * ((kCdd + ncoeffs[1] * kCd) * C +
                     (kSdd + ncoeffs[1] * kSd + ncoeffs[2]) * S)

        return K * kCd * sym.DiracDelta(t), G

    def ratfun(self, expr, s, t, **kwargs):

        if kwargs.pop('pdb', False):
            import pdb
            pdb.set_trace()

        sexpr = Ratfun(expr, s)

        if kwargs.get('damped_sin', False):
            if sexpr.degree == 2:
                return self.do_damped_sin(sexpr, s, t)
            # if False and sexpr.degree == 3 and Ratfun(expr * s).degree == 2:
            #    return self.do_damped_sin3(sexpr, s, t)

        self.debug('Finding QRPO representation')

        damping = kwargs.get('damping', None)
        Q, R, P, O, delay, undef = sexpr.as_QRPO(damping)

        if delay != 0:
            # This will be caught and trigger expansion of the expression.
            self.error('Unhandled delay %s' % delay)

        cresult = Zero

        if Q:
            Qpoly = sym.Poly(Q, s)
            C = Qpoly.all_coeffs()
            for n, c in enumerate(C):
                cresult += c * sym.diff(sym.DiracDelta(t), t, len(C) - n - 1)

        if R == []:
            return cresult, 0

        uresult = 0
        for m, (p, n, A) in enumerate(zip(P, O, R)):

            # This is zero for the conjugate pole.
            if R[m] == 0:
                continue

            # Search and remove conjugate pair.
            has_conjpair = False
            if p.is_complex and kwargs.get('pairs', True):
                pc = p.conjugate()
                Ac = A.conjugate()
                for m2, p2 in enumerate(P[m + 1:]):
                    m2 += m + 1
                    if n == O[m2] and p2 == pc and R[m2] == Ac:
                        R[m2] = 0
                        has_conjpair = True
                        break

            if has_conjpair:
                # Combine conjugate pairs.
                p = p.expand(complex=True)
                A = A.expand(complex=True)
                p_re = sym.re(p)
                p_im = sym.im(p)
                A_re = sym.re(A)
                A_im = sym.im(A)
                et = sym.exp(p_re * t)
                result = 2 * A_re * et * sym.cos(p_im * t)
                result -= 2 * A_im * et * sym.sin(p_im * t)
            else:
                result = A * sym.exp(p * t)

            if n > 1:
                result *= t ** (n - 1) / sym.factorial(n - 1)
            uresult += result

        # cresult is a sum of Dirac deltas and its derivatives so is known
        # to be causal.

        return cresult, uresult

    def product_undef(self, expr, s, t, **kwargs):

        # Handle expressions with a function of s, e.g., V(s) * Y(s), V(s)
        # / s etc.

        terms = expr.expand().as_ordered_terms()

        result = 0
        for term in terms:
            result += self.product_undef1(term, s, t, **kwargs)
        return result

    def product_undef1(self, expr, s, t, **kwargs):

        const, expr = factor_const(expr, s)

        factors = expr.as_ordered_factors()
        if len(factors) < 2:
            cresult, uresult = self.term1(expr, s, t, **kwargs)
            return const * (cresult + uresult)

        # Help s * 1 / (s + R * C) * I(s)
        if (len(factors) > 2 and not
            isinstance(factors[1], AppliedUndef) and
                isinstance(factors[2], AppliedUndef)):
            factors = [factors[0], factors[2], factors[1]] + factors[3:]

        if isinstance(factors[1], AppliedUndef):
            # Try to expose more simple cases, e.g. (R + s * L) * V(s)
            terms = factors[0].expand().as_ordered_terms()
            if len(terms) >= 2:
                result = Zero
                for term in terms:
                    result += self.product_undef(factors[1]
                                                 * term, s, t, **kwargs)
                return result * const

        cresult, uresult = self.term1(factors[0], s, t, **kwargs)
        result = cresult + uresult

        if kwargs.get('causal', False):
            # Assume that all functions are causal in the expression.
            t1 = Zero
            t2 = t
        else:
            # With unilateral Laplace transform need to set lower
            # limit at 0 rather than -oo.
            t1 = Zero
            t2 = sym.oo

        intnum = 0
        for m in range(len(factors) - 1):
            if m == 0 and isinstance(factors[1], AppliedUndef):
                # Note, as_ordered_factors puts powers of s before the functions.
                if factors[0] == s:
                    # Handle differentiation
                    # Convert s * V(s) to d v(t) / dt
                    result = self.func(factors[1], s, t)
                    result = sym.Derivative(result, t)
                    if not kwargs.get('zero_initial_conditions', True):
                        fname = factors[1].func.__name__
                        func = sym.Function(fname[0].lower() + fname[1:])
                        result += func(0) * sym.DiracDelta(t)
                    continue
                elif factors[0].is_Pow and factors[0].args[0] == s and factors[0].args[1] > 0:
                    # Handle higher order differentiation
                    # Convert s ** 2 * V(s) to d^2 v(t) / dt^2
                    result = self.func(factors[1], s, t)
                    result = sym.Derivative(result, t, factors[0].args[1])
                    if not kwargs.get('zero_initial_conditions', True):
                        fname = factors[1].func.__name__
                        func = sym.Function(fname[0].lower() + fname[1:])
                        v = func(t)
                        order = factors[0].args[1]
                        for m in range(order - 1):
                            result += sym.Derivative(v, t, m).subs(t, 0) * \
                                sym.DiracDelta(t, order - m - 1)
                        result += sym.Derivative(v, t, order - 1).subs(t, 0) * \
                            sym.DiracDelta(t)

                    continue
                elif factors[0].is_Pow and factors[0].args[0] == s \
                        and factors[0].args[1] == -1:
                    # Handle integration  1 / s * V(s)
                    tau = self.dummy_var(expr, 'tau', level=intnum, real=True)
                    intnum += 1
                    result = self.func(factors[1], s, tau)
                    result = sym.Integral(result, (tau, t1, t))
                    continue

            # Convert product to convolution
            tau = self.dummy_var(expr, 'tau', level=intnum, real=True)
            intnum += 1
            cresult, uresult = self.term1(factors[m + 1], s, t, **kwargs)
            expr2 = cresult + uresult
            result = sym.Integral(result.subs(t, t - tau) * expr2.subs(t, tau),
                                  (tau, t1, t2))

        return result * const

    def power(self, expr, s, t):

        # Handle expressions with a power of s.
        if not (expr.is_Pow and expr.args[0] == s):
            self.error('Expression not a power of s')
        exponent = expr.args[1]

        # Have many possible forms; the common ones are:
        # s**a, s**-a, s**(1+a), s**(1-a), s**-(1+a), s**(a-1)
        # Cannot tell if 1-a is positive.

        if exponent.is_positive:
            # Unfortunately, SymPy does not seem to support fractional
            # derivatives...
            return sym.Derivative(sym.DiracDelta(t), t, exponent,
                                  evaluate=False)

        if exponent.is_negative:
            return sym.Pow(t, -exponent - 1) / sym.Gamma(-exponent)

        self.error('Cannot determine sign of exponent')

    def delay_factor(self, expr, var):

        delay = Zero
        rest = One

        for f in expr.as_ordered_factors():
            b, e = f.as_base_exp()
            if b == sym.E and e.is_polynomial(var):
                p = sym.Poly(e, var)
                c = p.all_coeffs()
                if p.degree() == 1:
                    delay -= c[0]
                    if c[1] != 0:
                        rest *= sym.exp(c[1])
                    continue

            rest *= f
        return rest, delay

    def sympy(self, expr, s, t):

        self.debug('Resorting to SymPy')

        # This barfs when needing to generate Dirac deltas
        from sympy.integrals.transforms import inverse_laplace_transform
        result = inverse_laplace_transform(expr, s, t)

        if result.has(sym.InverseLaplaceTransform):
            self.error('SymPy does not know either')
        return result

    def tline_end(self, expr, s, t):
        """Attempt to find response at end of unterminated lossless
        transmission line."""

        # Look for expression of form 1 / (a * cosh(s * T) + b * sinh(s * T))

        func = sym.DiracDelta
        if (expr.is_Mul and expr.args[0].is_Pow and
                expr.args[0].args[0] == s and expr.args[0].args[1] == -1):
            expr *= s
            func = sym.Heaviside

        if not expr.is_Pow or expr.args[1] != -1:
            self.error('Not 1 / X')

        denom = expr.args[0]

        arg = None
        a = Zero
        b = Zero

        for term in denom.expand().as_ordered_terms():

            const, e = factor_const(term, s)
            if e.is_Function and e.func == sym.cosh:
                if arg is None:
                    arg = e.args[0]
                elif arg != e.args[0]:
                    self.error('Mismatch cosh arg')
                a += const
            elif e.is_Function and e.func == sym.sinh:
                if arg is None:
                    arg = e.args[0]
                elif arg != e.args[0]:
                    self.error('Mismatch sinh arg')
                b += const
            else:
                self.error('Term does not include cosh or sinh')

        T = arg / s
        if T.has(s):
            self.error('Frequency dependent delay')

        m = self.dummy_var(expr, 'm', level=0, real=True)
        g = (b - a) / (b + a)
        d = (b + a) / 2
        if g == 0:
            h = 1 / d * func(t - T)
        else:
            h = 1 / d * sym.Sum(g**m * func(t - (2 * m + 1)
                                * T), (m, 0, sym.oo))
        return h

    def tline_start(self, expr, s, t):
        """Attempt to find response at start of unterminated lossless
        transmission line."""

        # Look for expression of form
        # c * cosh(s * T) + d * sinh(s * T) / (a * cosh(s * T) + b * sinh(s * T))
        func = sym.DiracDelta
        if (expr.is_Mul and expr.args[0].is_Pow and
                expr.args[0].args[0] == s and expr.args[0].args[1] == -1):
            expr *= s
            func = sym.Heaviside

        numer, denom = expr.as_numer_denom()

        arg = None
        a = Zero
        b = Zero
        c = Zero
        d = Zero

        for term in denom.expand().as_ordered_terms():

            const, e = factor_const(term, s)
            if e.is_Function and e.func == sym.cosh:
                if arg is None:
                    arg = e.args[0]
                elif arg != e.args[0]:
                    self.error('Mismatch cosh arg')
                a += const
            elif e.is_Function and e.func == sym.sinh:
                if arg is None:
                    arg = e.args[0]
                elif arg != e.args[0]:
                    self.error('Mismatch sinh arg')
                b += const
            else:
                self.error('Term does not include cosh or sinh')

        for term in numer.expand().as_ordered_terms():

            const, e = factor_const(term, s)
            if e.is_Function and e.func == sym.cosh:
                if arg is None:
                    arg = e.args[0]
                elif arg != e.args[0]:
                    self.error('Mismatch cosh arg')
                c += const
            elif e.is_Function and e.func == sym.sinh:
                if arg is None:
                    arg = e.args[0]
                elif arg != e.args[0]:
                    self.error('Mismatch sinh arg')
                d += const
            else:
                self.error('Term does not include cosh or sinh')

        T = arg / s
        if T.has(s):
            self.error('Frequency dependent delay')

        Z0 = sym.sqrt((b * d - a * c) * d / (d**2 - c**2))
        Zl = Z0 * c / d
        Zs = (a - Z0 * Zl) / Z0
        K = d / Z0**2
        # The simplify can be slow
        Gammas = ((Zs - Z0) / (Zs + Z0)).simplify()
        Gammal = ((Zl - Z0) / (Zl + Z0)).simplify()

        m = self.dummy_var(expr, 'm', level=0, real=True)
        h1 = (1 - Gammas) / 2 * func(t)
        if Gammal == 0:
            h2 = 0
        elif Gammas == 0:
            h2 = (1 - Gammas) * (1 + Gammas) * func(t - 2 * m * T)
        else:
            h2 = (1 - Gammas) * (1 + Gammas) / Gammas * \
                sym.Sum((Gammas * Gammal)**m *
                        func(t - 2 * m * T), (m, 1, sym.oo))
        return K * (h1 + h2)

    def term1(self, expr, s, t, **kwargs):

        const, expr = factor_const(expr, s)

        if isinstance(expr, AppliedUndef):
            # Handle V(s), 3 * V(s) etc.  If causal is True it is assumed
            # that the unknown functions are causal.
            result = self.func(expr, s, t)
            return result * const, Zero

        if expr.has(AppliedUndef):
            return const * self.product_undef(expr, s, t, **kwargs), Zero

        try:
            # This is the common case.
            cresult, uresult = self.ratfun(expr, s, t, **kwargs)
            return const * cresult, const * uresult
        except:
            pass

        if expr.is_Pow and expr.args[1] == -1 and expr.args[0].is_Function:
            arg = expr.args[0].args[0]

            m = self.dummy_var(expr, 'm', level=0, real=True)
            if expr.args[0].func == sym.cosh:
                scale, shift = scale_shift(arg, s)
                if shift == 0:
                    return const * 2 * sym.Sum((-1)**m * sym.DiracDelta(t - scale * (2 * m + 1)), (m, 0, sym.oo)), Zero
            elif expr.args[0].func == sym.sinh:
                scale, shift = scale_shift(arg, s)
                if shift == 0:
                    return const * 2 * sym.Sum(sym.DiracDelta(t - scale * (2 * m + 1)), (m, 0, sym.oo)), Zero
            elif expr.args[0].func == sym.tanh:
                scale, shift = scale_shift(arg, s)
                if shift == 0:
                    return const * 2 * sym.Sum((-1)**m * sym.DiracDelta(t - scale * (2 * m + 1)), (m, 1, sym.oo)) + const * sym.DiracDelta(t), Zero

        if expr.has(sym.cosh) and expr.has(sym.sinh):
            try:
                return const * self.tline_end(expr, s, t), Zero
            except:
                pass
            try:
                return const * self.tline_start(expr, s, t), Zero
            except:
                pass

        if expr.is_Pow and expr.args[0] == s:
            return Zero, const * self.power(expr, s, t)

        self.error('Cannot determine inverse Laplace transform')

    def term(self, expr, s, t, **kwargs):

        expr, delay = self.delay_factor(expr, s)

        if delay == 0 and expr.has(sym.exp):
            # Handle cases like 1 / (s**2 * exp(5 * s) + s * exp(5 * s))
            expr1 = expr.simplify()
            expr2, delay2 = self.delay_factor(expr1, s)
            if not expr2.has(sym.exp):
                # Simplify can make things worse, e.g., 1 - exp(-5 *s)
                # becomes exp(-5 * s) * (exp(5 * s) - 1).  This needs
                # fixing in Sympy.
                expr = expr2
                delay = delay2

        try:
            cresult, uresult = self.term1(expr, s, t, **kwargs)
        except:

            # With deep=True, SymPy makes mess of (1 - exp(-s * T))
            terms = expr.expand(deep=False).as_ordered_terms()

            if len(terms) > 1:

                try:
                    # See if can convert to convolutions...
                    return self.product_undef(expr, s, t, **kwargs), Zero
                except:
                    pass

                uresult = Zero
                cresult = Zero

                for term in terms:
                    term = term.simplify()
                    cterm, uterm = self.term(term, s, t, **kwargs)
                    cresult += cterm
                    uresult += uterm
                return cresult, uresult

            expr = expr.simplify()

            try:
                cresult, uresult = self.term1(expr, s, t, **kwargs)
            except:
                return Zero, self.sympy(expr, s, t)

        if delay != 0:
            cresult = cresult.subs(t, t - delay)
            uresult = uresult.subs(t, t - delay)

            # h(t) = g(t - T)
            # If h(t) is known to be causal and T >= 0, then g(t)
            # is also causal.  If T > 0, then causality is violated.

            # If h(t) is not known to be causal then we can only infer
            # it for t >= 0 from its Laplace transform H(s).
            # From the delay theorem, H(s) = G(s) * exp(-s * T).
            # So given G(s) we can only infer g(t) for t >= 0.
            # This implies that we can only infer h(t) for t >= T.
            # This creates a can of worms since different terms of
            # an expression can have different conditions, say
            # exp(-3 * s) / s**2 + exp(-4 * s) / s.  So for now,
            # assume causal expression...

            if not kwargs.get('causal', False) and expr.has(AppliedUndef):
                warn('Assuming causal expression')

            if not delay.is_negative:
                if not delay.is_positive:
                    warn('Assuming %s is positive' % delay)
                cresult += uresult * sym.Heaviside(t - delay)
                uresult = Zero
            else:
                self.error('Causality violated with time advance %s.' % delay)

        else:
            if kwargs.get('causal', False):
                cresult += uresult * sym.Heaviside(t)
                uresult = Zero

        return cresult, uresult


inverse_laplace_transformer = InverseLaplaceTransformer()


def inverse_laplace_transform(expr, s, t, evaluate=True, **kwargs):
    """Calculate inverse Laplace of X(s) and return x(t).

    The unilateral Laplace transform cannot determine x(t) for n < 0
    unless given additional information:

    `dc` -- x(t) = constant
    `causal` -- x(t) = 0 for n < 0.
    `ac` -- x(t) = A cos(a * n) + B * sin(b * n)
    """

    return inverse_laplace_transformer.transform(expr, s, t,
                                                 evaluate=evaluate,
                                                 **kwargs)


def ILT(expr, s, t, evaluate=True, **kwargs):
    """Calculate inverse Laplace of X(s) and return x(t).

    The unilateral Laplace transform cannot determine x(t) for n < 0
    unless given additional information:

    `dc` -- x(t) = constant
    `causal` -- x(t) = 0 for n < 0.
    `ac` -- x(t) = A cos(a * n) + B * sin(b * n)
    """

    return inverse_laplace_transformer.transform(expr, s, t,
                                                 evaluate=evaluate,
                                                 **kwargs)



"""This module provides the SuperpositionCurrent class.  This
represents currents as a superposition in different transform domains.

For example, the following expression is a superposition of a DC
component, an AC component, and a transient component:

I1 = SuperpositionCurrent('1 + 2 * cos(2 * pi * 3 * t) + 3 * u(t)')

I1(t) returns the time domain expression
I1(s) returns the Laplace domain expression
I1(omega) returns the Fourier domain expression with angular frequency
I1(f) returns the Fourier domain expression with linear frequency

I1.dc returns the DC component
I1.ac returns a dictionary of the AC components, keyed by the frequency
I1.transient returns the time-domain transient component

I1.is_dc returns True if a pure DC signal
I1.is_ac returns True if a pure AC signal
I1.is_transient returns True if a pure transient signal
I1.has_dc returns True if has a DC signal
I1.has_ac returns True if has an AC signal
I1.has_transient returns True if has a transient signal

Copyright 2019--2023 Michael Hayes, UCECE

"""

from .super import Superposition
from .symbols import j, omega
from .impedance import impedance
from .currentmixin import CurrentMixin


class SuperpositionCurrent(Superposition, CurrentMixin):

    def cpt(self):
        from .oneport import I
        # Perhaps should generate more specific components such as Idc?
        return I(self.time())

    def _mul(self, x):
        if isinstance(x, (int, float)):
            return self.__scale__(x)

        if isinstance(x, Superposition):
            raise TypeError('Cannot multiply %s by %s. '
                            'You need to extract a specific component, e.g., a.s * b.s' %
                            (type(self).__name__, type(x).__name__))

        if not x.is_impedance:
            raise TypeError("Unsupported types for *: 'Current' and '%s'" %
                            type(x).__name__)
        if x.is_time_domain:
            raise TypeError("Cannot multiply by time-domain impedance.")

        obj = self
        if 't' in self and not x.is_constant_domain:
            obj = self.decompose()
        xs = x.laplace()

        new = SuperpositionVoltage()
        if 'dc' in obj:
            new += obj['dc'] * xs(0)
        for key in obj.ac_keys():
            new += obj[key] * xs(j * obj[key].omega)
        for key in obj.noise_keys():
            new += obj[key] * xs(omega)
        if 's' in obj:
            new += obj['s'] * xs
        if 't' in obj:
            new += obj['t'] * x
        return new

    def _div(self, x):
        if isinstance(x, (int, float)):
            return self.__scale__(1 / x)

        if isinstance(x, Superposition):
            raise TypeError("""
            Cannot divide %s by %s.  You need to extract a specific component, e.g., a.s / b.s.  If you want a transfer function use a(s) / b(s)""" % (type(self).__name__, type(x).__name__))

        if not x.is_admittance:
            raise TypeError("Cannot divide '%s' by '%s'; require admittance" %
                            (type(self).__name__, type(x).__name__))

        Z = 1 / x
        return self * Z

    def __mul__(self, x):
        if False:
            raise ValueError(
                'Cannot multiply superposition, need to convert to specific domain')
        return self._mul(x)

    def __rmul__(self, x):
        return self.__mul__(x)

    def __truediv__(self, x):
        if False:
            raise ValueError(
                'Cannot divide superposition, need to convert to specific domain')

        return self._div(x)


from .superpositionvoltage import SuperpositionVoltage  # nopep8



"""This module provides reactance support.

Copyright 2021--2022 Michael Hayes, UCECE

"""
from .expr import expr
from .sym import j
from .units import u as uu
from warnings import warn


def reactance(arg, **assumptions):
    """Create an impedance class for the specified reactance.

    The reactance, X, is defined as the imaginary part of the impedance:

    Z(omega) = R(omega) + j * X(omega)"""

    expr1 = expr(arg, frequency=True, **assumptions)
    if expr1.is_laplace_domain:
        warn('Specifying Laplace domain for reactance: %s' % expr1)

    if expr1.is_imaginary:
        warn('Reactance %s should be real' % expr1)

    expr1.units = uu.ohms
    return expr1



"""This module provides difference equation support.

Copyright 2021--2023 Michael Hayes, UCECE

"""

from .expr import equation, expr
from .nexpr import DiscreteTimeDomainExpression, nexpr
import sympy as sym


class DifferenceEquation(DiscreteTimeDomainExpression):

    def __init__(self, lhs, rhs, inputsym='x', outputsym='y', **assumptions):

        lhs = nexpr(lhs, **assumptions)
        rhs = nexpr(rhs, **assumptions)

        lhssymbols = lhs.symbols
        rhssymbols = rhs.symbols

        if inputsym not in rhssymbols:
            raise ValueError('Input symbol %s not in rhs %s' % (inputsym, rhs))
        if outputsym not in lhssymbols:
            raise ValueError('Output symbol %s not in lhs %s' %
                             (outputsym, lhs))

        cls = lhs.__class__
        eqn = sym.Eq(lhs.expr, rhs.expr, evaluate=False)
        super(DifferenceEquation, self).__init__(cls(eqn), **assumptions)
        self.inputsym = inputsym
        self.outputsym = outputsym

    def __eq__(self, x):

        if not isinstance(x, self.__class__):
            return False

        return self.lhs == x.lhs and self.rhs == x.rhs

    def transfer_function(self):
        """Create discrete-time transfer function."""
        from .zexpr import zexpr

        X = self.inputsym.upper()
        Y = self.outputsym.upper()

        result = self.ZT()

        # remove all initial conditions from z-transform if any
        for thisvar in [self.inputsym, self.outputsym]:
            counter = 0
            while result.has(sym.Function(thisvar)):
                result = result.subs(sym.Function(thisvar)(counter), 0)
                counter += 1

        # solve for Y(z) and remove X(z)
        result = result.solve(Y)[0] / zexpr(X + '(z)')

        #result = self.ZT().solve(Y)[0] / zexpr(X + '(z)')
        result.is_causal = True

        return result

    @property
    def lhs(self):
        """Return left-hand-side of difference equation."""

        return DiscreteTimeDomainExpression(self.expr.lhs, **self.assumptions)

    @property
    def rhs(self):
        """Return left-hand-side of difference equation."""

        return DiscreteTimeDomainExpression(self.expr.rhs, **self.assumptions)

    def dlti_filter(self):
        """Create discrete-time linear time-invariant filter."""

        return self.transfer_function().dlti_filter()

    def angular_frequency_response(self):
        """Return angular frequency response."""

        return self.transfer_function().angular_frequency_response()

    def frequency_response(self):
        """Return frequency response."""

        return self.transfer_function().frequency_response()

    def impulse_response(self):
        """Return impulse response."""

        return self.transfer_function().impulse_response()

    def step_response(self):
        """Return step response."""

        return self.transfer_function().step_response()

    def separate(self):
        """Rewrite difference equation so that input symbols are on the right
        and output symbols are on the left."""

        newlhs = 0
        newrhs = 0

        for term in self.lhs.as_ordered_terms():
            symbols = nexpr(term).symbols
            if self.outputsym in symbols:
                newlhs += term
            if self.inputsym in symbols:
                newrhs -= term

        for term in self.rhs.as_ordered_terms():
            symbols = nexpr(term).symbols
            if self.outputsym in symbols:
                newlhs -= term
            if self.inputsym in symbols:
                newrhs += term

        return self.__class__(newlhs, newrhs, **self.assumptions)


def difference_equation(lhs, rhs, inputsym='x', outputsym='y', **assumptions):
    """Create an Lcapy difference equation.

    This is an Lcapy expression of the form Eq(lhs, rhs).
    For example,
    e = difference_equation('y(n)', 'x(n) + 2 * y(n - 1)')

    The left hand side (lhs) and right hand side subexpressions
    can be obtained with the `lhs` and `rhs` attributes."""

    from .differenceequation import DifferenceEquation

    lhs = expr(lhs)
    rhs = expr(rhs)
    # Check if lhs and rhs compatible.
    diff = lhs - rhs

    return DifferenceEquation(lhs, rhs, inputsym, outputsym, **assumptions)



"""
This module implements an experimental time-stepping simulation.

Copyright 2021--2022 Michael Hayes, UCECE
"""

from .sym import tsym, symbol_map
from .symbols import oo
from warnings import warn

__all__ = ('Simulator', )

# The companion circuits use a Thevenin model.  This simplifies
# determination of the current through reactive components.  If the
# currents are not required, the Norton model would be faster since
# fewer nodes are needed and so the matrices are smaller.

# TODO:
# 1. handle initial values
# 2. offset correction


class SimulatedComponent(object):

    def __init__(self, cpt, v1_index, v2_index, v3_index, i_index):

        self.nodes = cpt.node_names
        self.name = cpt.name
        self.Reqname = 'R%seq' % cpt.name
        self.Veqname = 'V%seq' % cpt.name
        self.Reqsym = symbol_map(self.Reqname)
        self.Veqsym = symbol_map(self.Veqname)
        self.v1_index = v1_index
        self.v2_index = v2_index
        # This is the dummy node required for the Thevenin companion circuit.
        self.v3_index = v3_index
        self.i_index = i_index

    def subsdict(self, n, dt, v1, v2, i):
        """Create a dictionary of substitutions."""

        geq = self.geq(n, dt, v1, v2, i)
        veq = self.veq(n, dt, v1, v2, i)

        return {self.Reqsym: 1 / geq, self.Veqsym: veq}

    def stamp(self, A, Z, num_nodes, n, dt, v1, v2, i):

        geq = self.geq(n, dt, v1, v2, i)
        veq = self.veq(n, dt, v1, v2, i)

        n1, n2 = self.v1_index, self.v3_index

        if n1 >= 0 and n2 >= 0:
            A[n1, n2] -= geq
            A[n2, n1] -= geq
        if n1 >= 0:
            A[n1, n1] += geq
        if n2 >= 0:
            A[n2, n2] += geq

        m = self.i_index + num_nodes
        Z[m] += veq


class SimulatedCapacitor(SimulatedComponent):

    def __init__(self, C, v1_index, v2_index, v3_index, i_index):

        super(SimulatedCapacitor, self).__init__(C, v1_index, v2_index,
                                                 v3_index, i_index)
        self.Cval = C.C.expr


class SimulatedInductor(SimulatedComponent):

    def __init__(self, L, v1_index, v2_index, v3_index, i_index):

        super(SimulatedInductor, self).__init__(L, v1_index, v2_index,
                                                v3_index, i_index)
        self.Lval = L.L.expr


class SimulatedCapacitorTrapezoid(SimulatedCapacitor):

    def geq(self, n, dt, v1, v2, i):

        return (2 * self.Cval) / dt

    def veq(self, n, dt, v1, v2, i):

        if n < 1:
            return 0

        v = v1[n - 1] - v2[n - 1]

        geq = (2 * self.Cval) / dt
        veq = v + i[n - 1] / geq
        return veq


class SimulatedInductorTrapezoid(SimulatedInductor):

    def geq(self, n, dt, v1, v2, i):

        return dt / (2 * self.Lval)

    def veq(self, n, dt, v1, v2, i):

        if n < 1:
            return 0

        v = v1[n - 1] - v2[n - 1]

        geq = dt / (2 * self.Lval)
        veq = -v - i[n - 1] / geq
        return veq


class SimulatedCapacitorBackwardEuler(SimulatedCapacitor):

    def geq(self, n, dt, v1, v2, i):

        return self.Cval / dt

    def veq(self, n, dt, v1, v2, i):

        if n < 1:
            return 0

        return v1[n - 1] - v2[n - 1]


class SimulatedInductorBackwardEuler(SimulatedInductor):

    def geq(self, n, dt, v1, v2, i):

        return dt / self.Lval

    def veq(self, n, dt, v1, v2, i):

        geq = dt / self.Lval
        veq = -i[n - 1] / geq
        return veq


class SimulationResultsNode(object):

    def __init__(self, v):
        self.v = v


class SimulationResultsCpt(object):

    def __init__(self, v, i):
        self.v = v
        self.i = i


class SimulationResults(object):

    def __init__(self, tv, cct, r_model, node_list, branch_list):

        from numpy import zeros

        self.t = tv
        self.cct = cct
        self.r_model = r_model

        N = len(tv)

        # MNA calculates the node voltages plus currents through
        # inductors and voltage sources.  The currents through the
        # other elements can be found from the voltage difference
        # divided by the element resistance.

        self.num_nodes = len(node_list) - 1
        self.num_branches = len(branch_list)

        self.node_voltages = zeros((self.num_nodes + 1, N))
        self.branch_currents = zeros((self.num_branches, N))

    def __getitem__(self, name):
        """Return element or node by name."""

        cct = self.cct

        # If name is an integer, convert to a string.
        if isinstance(name, int):
            name = '%d' % name

        if name in cct.nodes:
            return SimulationResultsNode(self._node_voltage_get(name))

        if name in cct._elements:
            return SimulationResultsCpt(self._cpt_voltage_get(name),
                                        self._cpt_current_get(name))

        raise AttributeError('Unknown element or node name %s' % name)

    def __getattr__(self, attr):
        """Return element or node by name.
        This gets called if there is no explicit attribute attr for
        this instance.  This is primarily for accessing elements and
        non-numerical node names.  It also gets called if the called
        attr throws an AttributeError exception.  The annoying thing
        is that hasattr uses getattr and checks for an exception.

        """

        return self.__getitem__(attr)

    def _node_voltage_get(self, n):

        index = self.r_model.mna._node_index(n)
        # NB, node_voltages is zero for index = -1
        return self.node_voltages[index]

    def _cpt_voltage_get(self, cptname):

        cpt = self.cct.elements[cptname]

        v1 = self._node_voltage_get(cpt.node_names[0])
        v2 = self._node_voltage_get(cpt.node_names[1])
        return v1 - v2

    def _cpt_current_get(self, cptname):

        try:
            index = self.r_model.mna._branch_index(cptname)
            return self.branch_currents[index]
        except:
            cpt = self.cct._elements[cptname]
            if cpt.is_capacitor or cpt.is_inductor:
                # For a capacitor we can find the current through the
                # companion resistor or voltage source.
                return self._cpt_current_get('V%seq' % cptname)

            Vd = self._cpt_voltage_get(cptname)
            if cpt.is_resistor or cpt.is_conductor:
                return Vd / float(cpt.R.expr)

            # Need to determine resistance of the cpt
            raise ValueError('FIXME')

    @property
    def V(self, node):
        """Node voltage with respect to ground."""

        return self._node_voltage_get(node)


class Simulator(object):

    def __init__(self, cct):
        """Create simulation object for the circuit specified by `cct`.

        All the symbolic circuit component values need to be replaced
        with numerical values (using the subs method) except for
        functions of t, such as Heaviside(t).

        Here's an example of use:

        cct = Circuit('circuit.sch')
        sim = Simulator(cct)
        t = np.linspace(0, 1, 100)
        results = sim(t)

        plot(t, results.C1.v)

        """

        self.cct = cct

        # Companion resistor model
        self.r_model = cct.r_model().subcircuits['time']

    def _step(self, foo, n, tv, results):

        from numpy import array, linalg, dot

        # Substitute values into the MNA A matrix and Z vector,
        # then perform numerical inversion of the A matrix.
        # Alternatively, a faster way, if the A matrix was small,
        # would be to symbolically invert the A matrix and
        # then substitute values.

        if n == 0:
            if not self.cct.is_IVP:
                # Initial voltages and currents all zero.
                return
            p_model = self.cct.pre_initial_model()
            # Evaluate model and copy node voltages and branch currents...

            return

        dt = tv[n] - tv[n - 1]
        subsdict = {tsym: tv[n]}

        Zsym = self.Zsym
        Zsym = Zsym.subs(subsdict)
        Z = array(Zsym).astype(float).squeeze()

        if n == 1 and Zsym.free_symbols != set():
            raise ValueError(
                'Undefined symbols %s in Z vector; use subs to replace with numerical values' % Zsym.free_symbols)

        # Ensure have a copy.
        A = self.A + 0

        for cpt in self.reactive_cpts:

            # NB, node_voltages is zero for index = -1
            v1 = results.node_voltages[cpt.v1_index]
            v2 = results.node_voltages[cpt.v2_index]
            i = results.branch_currents[cpt.i_index]

            cpt.stamp(A, Z, results.num_nodes, n, dt, v1, v2, i)

        Ainv = linalg.inv(A)

        results1 = dot(Ainv, Z)

        num_nodes = results.num_nodes
        results.node_voltages[0:num_nodes, n] = results1[0:num_nodes]
        results.branch_currents[:, n] = results1[num_nodes:]

    def __call__(self, tv, integrator='trapezoid'):
        """Numerically evaluate circuit using time-stepping numerical
        integration at the vector of times specified by `tv`.

        Currently the only supported integration methods are
        trapezoidal and backward-euler (others would be trivial to
        add).  The trapezoidal integration method is the default since
        it is accurate but it can be unstable producing some
        oscillations.  Unfortunately, there is no ideal numerical
        integration method and there is always a tradeoff between
        accuracy and stability.

        """

        from numpy import array

        if integrator == 'trapezoid':
            Ccls = SimulatedCapacitorTrapezoid
            Lcls = SimulatedInductorTrapezoid
        elif integrator == 'backward-euler':
            Ccls = SimulatedCapacitorBackwardEuler
            Lcls = SimulatedInductorBackwardEuler
        else:
            raise ValueError('Unknown integrator ' + integrator)

        r_model = self.r_model

        Asubsdict = {}
        Zsubsdict = {}
        self.reactive_cpts = []
        for key, elt in self.cct.elements.items():
            if not (elt.is_inductor or elt.is_capacitor):
                continue

            if not elt.has_ic:
                warn('Initial conditions for %s ignored' % elt.name)

            v1_index = r_model.mna._node_index(elt.node_names[0])
            v2_index = r_model.mna._node_index(elt.node_names[1])
            i_index = r_model.mna._branch_index('V%seq' % elt.name)
            relt = self.r_model.elements['R%seq' % elt.name]
            v3_index = r_model.mna._node_index(relt.node_names[1])

            if elt.is_inductor:
                cls = Lcls
            else:
                cls = Ccls

            simcpt = cls(elt, v1_index, v2_index, v3_index, i_index)
            self.reactive_cpts.append(simcpt)

            Asubsdict[simcpt.Reqsym] = oo
            Zsubsdict[simcpt.Veqsym] = 0

        # Remove 1 / Req entries
        Asym = r_model.mna._A.subs(Asubsdict)
        # Remove Veq entries
        Zsym = r_model.mna._Z.subs(Zsubsdict)

        self.Asym = Asym
        self.Zsym = Zsym

        if Asym.free_symbols != set():
            raise ValueError(
                'Undefined symbols %s in A matrix; use subs to replace with numerical values' % Asym.free_symbols)

        # Convert to numpy ndarray
        self.A = array(Asym).astype(float)

        results = SimulationResults(tv, self.cct, r_model,
                                    r_model.node_list,
                                    r_model.mna.unknown_branch_currents)

        for n, t1 in enumerate(tv):
            self._step(r_model, n, tv, results)

        return results



"""This module provides the TransformDomains class.

Copyright 2023 Michael Hayes, UCECE

"""


from .expr import Expr
from .symbols import omega


class TransformDomains(dict):

    def __getattr__(self, attr):
        if attr not in self:
            raise AttributeError('Unknown attribute %s' % attr)
        return self[attr]

    def __getitem__(self, key):
        if key == 'w':
            key = omega
        # This allows a[omega] to work if omega used as key
        # instead of 'omega'.
        if isinstance(key, Expr):
            key = key.expr
        return super(TransformDomains, self).__getitem__(key)



"""This module provides the Context class that maintains the context
for a Circuit.

Copyright 2014--2022 Michael Hayes, UCECE

"""


class Context(object):

    def __init__(self):
        self.assumptions = {}
        # Noise instance identifier
        self.nid = 0



"""
This module supports simple linear one-port networks based on the
following ideal components:

V independent voltage source
I independent current source
R resistor
C capacitor
L inductor

These components are converted to s-domain models and so capacitor and
inductor components can be specified with initial voltage and
currents, respectively, to model transient responses.

One-ports can either be connected in series (+) or parallel (|) to
create a new one-port.

Copyright 2014--2022 Michael Hayes, UCECE
"""

from __future__ import division
from .functions import Heaviside, cos, exp, Derivative, Integral
from .sym import omega0sym, tsym, ksym, oo
from .symbols import j, t, s
from .network import Network
from .immittancemixin import ImmittanceMixin
from .impedance import impedance
from .admittance import admittance
from .voltage import voltage
from .current import current
from warnings import warn


__all__ = ('V', 'I', 'v', 'i', 'R', 'NR', 'L', 'C', 'G', 'Y', 'Z',
           'Vdc', 'Vstep', 'Idc', 'Istep', 'Vac', 'sV', 'sI',
           'Iac', 'Vnoise', 'Inoise',
           'Par', 'Ser', 'Xtal', 'FerriteBead', 'CPE', 'series', 'parallel',
           'ladder')


def _check_oneport_args(args):

    for arg1 in args:
        if not isinstance(arg1, OnePort):
            raise ValueError('%s not a OnePort' % arg1)


class OnePort(Network, ImmittanceMixin):
    """One-port network

    There are four major types of OnePort:
       VoltageSource
       CurrentSource
       Impedance
       Admittance
       ParSer for combinations of OnePort

    Attributes: Y, Z, Voc, Isc, y, z, voc, isc
      Y = Y(s)  admittance
      Z = Z(s)  impedance
      Voc       open-circuit voltage in appropriate transform domain
      Isc       short-circuit current in appropriate transform domain
      y = y(t)  impulse response of admittance
      z = z(t)  impulse response of impedance
      voc = voc(t) open-circuit voltage time response
      isc = isc(t) short-circuit current time response
    """

    # Dimensions and separations of component with horizontal orientation.
    height = 0.3
    hsep = 0.5
    width = 1
    wsep = 0.25

    _Z = None
    _Y = None
    _Voc = None
    _Isc = None

    @property
    def impedance(self):
        if self._Z is not None:
            return self._Z
        if self._Y is not None:
            return 1 / self._Y
        if self._Voc is not None:
            return impedance(0)
        if self._Isc is not None:
            return 1 / admittance(0)
        raise ValueError('_Isc, _Voc, _Y, or _Z undefined for %s' % self)

    @property
    def admittance(self):
        if self._Y is not None:
            return self._Y
        return 1 / self.impedance

    @property
    def Voc(self):
        """Open-circuit voltage."""
        if self._Voc is not None:
            return self._Voc
        if self._Isc is not None:
            return self._Isc._mul(self.impedance)
        if self._Z is not None or self._Y is not None:
            return SuperpositionVoltage(0)
        raise ValueError('_Isc, _Voc, _Y, or _Z undefined for %s' % self)

    @property
    def Isc(self):
        """Short-circuit current.  This is the current flowing out
        of the positive node of the network, through a wire, and back
        to the negative node of the network."""
        if self._Isc is not None:
            return self._Isc
        return self.Voc._mul(self.admittance)

    @property
    def V(self):
        """Open-circuit voltage."""
        return self.Voc

    @property
    def I(self):
        """Open-circuit current.  Except for a current source this is zero."""
        return SuperpositionCurrent(0)

    @property
    def i(self):
        """Open-circuit time-domain current.  Except for a current source this
        is zero."""
        return self.I.time()

    def __add__(self, OP):
        """Series combination"""

        return Ser(self, OP)

    def __or__(self, OP):
        """Parallel combination"""

        return Par(self, OP)

    @property
    def has_series_I(self):
        return self.is_current_source

    @property
    def has_parallel_V(self):
        return self.is_voltage_source

    def chain(self, TP):
        """Chain to a two-port.  This is experimental."""

        if isinstance(TP, OnePort):
            raise ValueError('Cannot chain oneport with a oneport')

        from .twoport import TwoPort
        if not isinstance(TP, TwoPort):
            raise ValueError('%s not a twoport' % TP)

        return TP.source(self)

    def series(self, OP):
        """Series combination"""

        return Ser(self, OP)

    def parallel(self, OP):
        """Parallel combination"""

        return Par(self, OP)

    def ladder(self, *args):
        """Create (unbalanced) ladder network"""

        return Ladder(self, *args)

    def lsection(self, OP2):
        """Create L section (voltage divider)"""

        if not issubclass(OP2.__class__, OnePort):
            raise TypeError('Argument not ', OnePort)

        return LSection(self, OP2)

    def tsection(self, OP2, OP3):
        """Create T section"""

        if not issubclass(OP2.__class__, OnePort):
            raise TypeError('Argument not ', OnePort)

        if not issubclass(OP3.__class__, OnePort):
            raise TypeError('Argument not ', OnePort)

        return TSection(self, OP2, OP3)

    def expand(self):

        return self

    def load(self, OP2):

        if not issubclass(OP2.__class__, OnePort):
            raise TypeError('Load argument not ', OnePort)

        return LoadCircuit(self, OP2)

    @property
    def voc(self):
        """Open-circuit time-domain voltage."""
        return self.Voc.time()

    @property
    def isc(self):
        """Short-circuit time-domain current.  This is the current flowing out
        of the positive node of the network, through a wire, and back
        to the negative node of the network."""
        return self.Isc.time()

    @property
    def v(self):
        """Open-circuit time-domain voltage."""
        return self.voc

    @property
    def z(self):
        """Impedance impulse-response."""
        return self.impedance.time()

    @property
    def y(self):
        """Admittance impulse-response."""
        return self.admittance.time()

    def thevenin(self):
        """Simplify to a Thevenin network"""

        new = self.simplify()
        Voc = new.Voc
        Z = new.impedance

        if Voc.is_superposition and not Z.is_real:
            warn('Detected superposition with reactive impedance, using s-domain.')
            Z1 = Z
            V1 = Voc.laplace()
        elif Voc.is_ac:
            Z1 = Z.subs(j * Voc.ac_keys()[0])
            V1 = Voc.select(Voc.ac_keys()[0])
        elif Voc.is_dc:
            Z1 = Z.subs(0)
            V1 = Voc(0)
        else:
            V1 = Voc
            Z1 = Z

        V1 = V1.cpt()
        Z1 = Z1.cpt()

        if Voc == 0:
            return Z1
        if Z == 0:
            return V1

        return Ser(Z1, V1)

    def norton(self):
        """Simplify to a Norton network"""

        new = self.simplify()
        Isc = new.Isc
        Y = new.admittance

        if Isc.is_superposition and not Y.is_real:
            warn('Detected superposition with reactive impedance, using s-domain.')
            Y1 = Y
            I1 = Isc.laplace()
        elif Isc.is_ac:
            Y1 = Y.subs(j * Isc.ac_keys()[0])
            I1 = Isc.select(Isc.ac_keys()[0])
        elif Isc.is_dc:
            Y1 = Y.subs(0)
            I1 = Isc(0)
        else:
            I1 = Isc
            Y1 = Y

        I1 = I1.cpt()
        Y1 = Y1.cpt()

        if Isc == 0:
            return Y1
        if Y == 0:
            return I1

        return Par(Y1, I1)

    def s_model(self):
        """Convert to s-domain."""

        if self._Voc is not None:
            if self._Voc == 0:
                return Z(self.impedance)
            Voc = self._Voc.laplace()
            if self.Z == 0:
                return V(Voc)
            return Ser(V(Voc), Z(self.impedance))
        elif self._Isc is not None:
            if self._Isc == 0:
                return Y(self.admittance)
            Isc = self._Isc.laplace()
            if self.admittance == 0:
                return I(Isc)
            return Par(I(Isc), Y(self.admittance))
        elif self._Z is not None:
            return Z(self._Z)
        elif self._Y is not None:
            return Y(self._Y)
        raise RuntimeError('Internal error')

    def noise_model(self):
        """Convert to noise model."""

        from .symbols import omega

        if not isinstance(self, (R, G, Y, Z)):
            return self

        R1 = self.R
        if R1 != 0:
            Vn = Vnoise('sqrt(4 * k_B * T * %s)' % R1(j * omega))
            return self + Vn
        return self

        def current_equation(self, v, kind='t'):
            """Return expression for current through component given
            applied voltage."""

            raise NotImplementedError('current_equation not defined')

        def voltage_equation(self, i, kind='t'):
            """Return expression for voltage across component given
            applied current."""

            raise NotImplementedError('voltage_equation not defined')

    def _Zkind(self, kind):

        # This is for determining impedances
        if not isinstance(kind, str):
            # AC
            domain = kind
        elif kind in ('super', 'time', 't'):
            domain = 't'
        elif kind in ('laplace', 'ivp', 's'):
            domain = 's'
        elif kind == 'dc':
            domain = 0
        elif kind.startswith('n'):
            domain = 'f'
        else:
            raise RuntimeError('Unhandled circuit kind ' + kind)
        return self.Z(domain)


class ParSer(OnePort):
    """Parallel/serial class"""

    def __str__(self):

        str = ''

        for m, arg in enumerate(self.args):
            argstr = arg.__str__()

            if isinstance(arg, ParSer) and arg.__class__ != self.__class__:
                argstr = '(' + argstr + ')'

            str += argstr

            if m != len(self.args) - 1:
                str += ' %s ' % self._operator

        return str

    def _repr_pretty_(self, p, cycle):

        p.text(self.pretty())

    def _repr_latex_(self):

        return '$%s$' % self.latex()

    def pretty(self, **kwargs):

        str = ''

        for m, arg in enumerate(self.args):
            argstr = arg.pretty(**kwargs)

            if isinstance(arg, ParSer) and arg.__class__ != self.__class__:
                argstr = '(' + argstr + ')'

            str += argstr

            if m != len(self.args) - 1:
                str += ' %s ' % self._operator

        return str

    def pprint(self):

        print(self.pretty())

    def latex(self):

        str = ''

        for m, arg in enumerate(self.args):
            argstr = arg.latex()

            if isinstance(arg, ParSer) and arg.__class__ != self.__class__:
                argstr = '(' + argstr + ')'

            str += argstr

            if m != len(self.args) - 1:
                str += ' %s ' % self._operator

        return str

    def _combine(self, arg1, arg2):

        if arg1.__class__ != arg2.__class__:
            if self.__class__ == Ser:
                if isinstance(arg1, V) and arg1.Voc == 0:
                    return arg2
                if isinstance(arg2, V) and arg2.Voc == 0:
                    return arg1
                if isinstance(arg1, (R, Z)) and arg1.impedance == 0:
                    return arg2
                if isinstance(arg2, (R, Z)) and arg2.impedance == 0:
                    return arg1
            if self.__class__ == Par:
                if isinstance(arg1, I) and arg1.Isc == 0:
                    return arg2
                if isinstance(arg2, I) and arg2.Isc == 0:
                    return arg1
                if isinstance(arg1, (Y, G)) and arg1.admittance == 0:
                    return arg2
                if isinstance(arg2, (Y, G)) and arg2.admittance == 0:
                    return arg1

            return None

        if self.__class__ == Ser:
            if isinstance(arg1, I):
                return None
            if isinstance(arg1, Vdc):
                return Vdc(arg1.v0 + arg2.v0)
            # Could simplify Vac here if same frequency
            if isinstance(arg1, V):
                return V(arg1 + arg2)
            if isinstance(arg1, R):
                return R(arg1._R + arg2._R)
            if isinstance(arg1, L):
                # The currents should be the same!
                if arg1.i0 != arg2.i0 or arg1.has_ic != arg2.has_ic:
                    raise ValueError('Series inductors with different'
                                     ' initial currents!')
                i0 = arg1.i0 if arg1.has_ic else None
                return L(arg1.L + arg2.L, i0)
            if isinstance(arg1, G):
                return G(arg1._G * arg2._G / (arg1._G + arg2._G))
            if isinstance(arg1, C):
                v0 = arg1.v0 + arg2.v0 if arg1.has_ic or arg2.has_ic else None
                return C(arg1.C * arg2.C / (arg1.C + arg2.C), v0)
            return None

        elif self.__class__ == Par:
            if isinstance(arg1, V):
                return None
            if isinstance(arg1, Idc):
                return Idc(arg1.i0 + arg2.i0)
            # Could simplify Iac here if same frequency
            if isinstance(arg1, I):
                return I(arg1 + arg2)
            if isinstance(arg1, G):
                return G(arg1._G + arg2._G)
            if isinstance(arg1, C):
                # The voltages should be the same!
                if arg1.v0 != arg2.v0 or arg1.has_ic != arg2.has_ic:
                    raise ValueError('Parallel capacitors with different'
                                     ' initial voltages!')
                v0 = arg1.v0 if arg1.has_ic else None
                return C(arg1.C + arg2.C, v0)
            if isinstance(arg1, R):
                return R(arg1._R * arg2._R / (arg1._R + arg2._R))
            if isinstance(arg1, L):
                i0 = arg1.i0 + arg2.i0 if arg1.has_ic or arg2.has_ic else None
                return L(arg1.L * arg2.L / (arg1.L + arg2.L), i0)
            return None

        else:
            raise TypeError('Undefined class')

    def simplify(self, deep=True):
        """Perform simple simplifications, such as parallel resistors,
        series inductors, etc., rather than collapsing to a Thevenin
        or Norton network.

        This does not expand compound components such as crystal
        or ferrite bead models.  Use expand() first.
        """

        # Simplify args (recursively) and combine operators if have
        # Par(Par(A, B), C) etc.
        new = False
        newargs = []
        for m, arg in enumerate(self.args):
            if isinstance(arg, ParSer):
                arg = arg.simplify(deep)
                new = True
                if arg.__class__ == self.__class__:
                    newargs.extend(arg.args)
                else:
                    newargs.append(arg)
            else:
                newargs.append(arg)

        if new:
            self = self.__class__(*newargs)

        # Scan arg list looking for compatible combinations.
        # Could special case the common case of two args.
        new = False
        args = list(self.args)
        for n in range(len(args)):

            arg1 = args[n]
            if arg1 is None:
                continue
            if isinstance(arg1, ParSer):
                continue

            for m in range(n + 1, len(args)):

                arg2 = args[m]
                if arg2 is None:
                    continue
                if isinstance(arg2, ParSer):
                    continue

                # TODO, think how to simplify things such as
                # Par(Ser(V1, R1), Ser(R2, V2)).
                # Could do Thevenin/Norton transformations.

                newarg = self._combine(arg1, arg2)
                if newarg is not None:
                    # print('Combining', arg1, arg2, 'to', newarg)
                    args[m] = None
                    arg1 = newarg
                    new = True

            args[n] = arg1

        if new:
            args = [arg for arg in args if arg is not None]
            if len(args) == 1:
                return args[0]
            self = self.__class__(*args)

        return self

    def expand(self):
        """Expand compound components such as crystals or ferrite bead
        models into R, L, G, C, V, I"""

        newargs = []
        for m, arg in enumerate(self.args):
            newarg = arg.expand()
            newargs.append(newarg)

        return self.__class__(*newargs)

    def s_model(self):
        """Convert to s-domain."""
        args = [arg.s_model() for arg in self.args]
        return (self.__class__(*args))

    def noise_model(self):
        """Convert to noise model."""
        args = [arg.noise_model() for arg in self.args]
        return (self.__class__(*args))

    @property
    def Isc(self):
        return self.cct.Isc(1, 0)

    @property
    def Voc(self):
        return self.cct.Voc(1, 0)


class Par(ParSer):
    """Parallel class"""

    _operator = '|'
    is_parallel = True

    def __init__(self, *args):

        if len(args) < 2:
            raise ValueError('Par requires at least two args')

        _check_oneport_args(args)
        super(Par, self).__init__()
        self.args = args

        for n, arg1 in enumerate(self.args):
            for arg2 in self.args[n + 1:]:
                if isinstance(arg1, V) and isinstance(arg2, V):
                    raise ValueError('Voltage sources connected in parallel'
                                     ' %s and %s' % (arg1, arg2))
                elif isinstance(arg1, V):
                    print('Warn: redundant component %s in parallel with voltage source %s' % (
                        arg2, arg1))

                elif isinstance(arg2, V):
                    print('Warn: redundant component %s in parallel with voltage source %s' % (
                        arg1, arg2))

    @property
    def width(self):

        total = 0
        for arg in self.args:
            val = arg.width
            if val > total:
                total = val
        return total + 2 * self.wsep

    @property
    def height(self):

        total = 0
        for arg in self.args:
            total += arg.height
        return total + (len(self.args) - 1) * self.hsep

    def _net_make(self, netlist, n1=None, n2=None, dir='right'):

        s = []
        if n1 is None:
            n1 = netlist._node
        n3, n4 = netlist._node, netlist._node

        H = [(arg.height + self.hsep) * 0.5 for arg in self.args]

        N = len(H)
        num_branches = N // 2

        # Draw component in centre if have odd number in parallel.
        if N & 1:
            s.append(self.args[N // 2]._net_make(netlist, n3, n4, dir))

        na, nb = n3, n4

        s.append('W %s %s; %s=%s' % (n1, n3, dir, self.wsep))

        if dir == 'right':
            updir, downdir = 'up', 'down'
        else:
            updir, downdir = 'right', 'left'

        # Draw components above centre
        for n in range(num_branches):

            if not (N & 1) and n == 0:
                sep = H[N // 2 - 1]
            else:
                sep = H[N // 2 - n] + H[N // 2 - 1 - n]

            nc, nd = netlist._node, netlist._node
            s.append('W %s %s; %s=%s' % (na, nc, updir, sep))
            s.append('W %s %s; %s=%s' % (nb, nd, updir, sep))
            s.append(self.args[N // 2 - 1 - n]._net_make(netlist, nc, nd, dir))
            na, nb = nc, nd

        na, nb = n3, n4

        # Draw components below centre
        for n in range(num_branches):

            if not (N & 1) and n == 0:
                sep = H[(N + 1) // 2]
            else:
                sep = H[(N + 1) // 2 + n] + H[(N + 1) // 2 - 1 + n]

            nc, nd = netlist._node, netlist._node
            s.append('W %s %s; %s=%s' % (na, nc, downdir, sep))
            s.append('W %s %s; %s=%s' % (nb, nd, downdir, sep))
            s.append(self.args[(N + 1) // 2 +
                     n]._net_make(netlist, nc, nd, dir))
            na, nb = nc, nd

        if n2 is None:
            n2 = netlist._node

        s.append('W %s %s; %s=%s' % (n4, n2, dir, self.wsep))
        return '\n'.join(s)

    @property
    def has_parallel_V(self):

        for cpt1 in self.args:
            if cpt1.has_parallel_V:
                return True
        return False

    @property
    def admittance(self):
        Y = 0
        for arg in self.args:
            Y += arg.admittance
        return Y

    @property
    def impedance(self):
        return 1 / self.admittance

    @property
    def Isc(self):
        I = 0
        for arg in self.args:
            I += arg.Isc
        return I


class Ser(ParSer):
    """Series class"""

    _operator = '+'
    is_series = True

    def __init__(self, *args):

        if len(args) < 2:
            raise ValueError('Ser requires at least two args')

        _check_oneport_args(args)
        super(Ser, self).__init__()
        self.args = args

        for n, arg1 in enumerate(self.args):
            for arg2 in self.args[n + 1:]:
                if isinstance(arg1, I) and isinstance(arg2, I):
                    raise ValueError('Current sources connected in series'
                                     ' %s and %s' % (arg1, arg2))
                elif isinstance(arg1, I):
                    print(
                        'Warn: redundant component %s in series with current source %s' % (arg2, arg1))

                elif isinstance(arg2, I):
                    print(
                        'Warn: redundant component %s in series with current source %s' % (arg1, arg2))

    @property
    def height(self):

        total = 0
        for arg in self.args:
            val = arg.height
            if val > total:
                total = val
        return total

    @property
    def width(self):

        total = 0
        for arg in self.args:
            total += arg.width
        return total + (len(self.args) - 1) * self.wsep

    def _net_make(self, netlist, n1=None, n2=None, dir='right'):

        s = []
        if n1 is None:
            n1 = netlist._node
        for arg in self.args[:-1]:
            n3 = netlist._node
            s.append(arg._net_make(netlist, n1, n3, dir))
            n1 = netlist._node
            s.append('W %s %s; %s=%s' % (n3, n1, dir, self.wsep))

        if n2 is None:
            n2 = netlist._node
        s.append(self.args[-1]._net_make(netlist, n1, n2, dir))
        return '\n'.join(s)

    @property
    def has_series_I(self):
        for cpt1 in self.args:
            if cpt1.has_series_I:
                return True
        return False

    @property
    def admittance(self):
        return 1 / self.impedance

    @property
    def impedance(self):
        Z = 0
        for arg in self.args:
            Z += arg.impedance
        return Z

    @property
    def Voc(self):
        V = 0
        for arg in self.args:
            V += arg.Voc
        return V


class R(OnePort):
    """Resistor"""

    is_resistor = True
    is_noiseless = False

    def __init__(self, Rval='R', **kwargs):

        self.kwargs = kwargs
        self.args = (Rval, )
        self._R = cexpr(Rval)
        self._Z = impedance(self._R, causal=True)

    def current_equation(self, v, kind='t'):
        """Return expression for current through component given
        applied voltage."""

        return SuperpositionCurrent(SuperpositionVoltage(v).select(kind) / self._Z).select(kind)

    def voltage_equation(self, i, kind='t'):
        """Return expression for voltage across component given
        applied current."""

        return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Z).select(kind)


class NR(R):
    """Noiseless resistor"""

    is_noiseless = True


class G(OnePort):
    """Conductor"""

    is_conductor = True
    is_noiseless = False

    def __init__(self, Gval='G', **kwargs):

        self.kwargs = kwargs
        self.args = (Gval, )
        self._G = cexpr(Gval)
        self._Z = impedance(1 / self._G, causal=True)

    def _net_make(self, netlist, n1=None, n2=None, dir='right'):

        if n1 == None:
            n1 = netlist._node
        if n2 == None:
            n2 = netlist._node
        opts_str = self._opts_str(dir)
        return 'R? %s %s {%s}; %s' % (n1, n2, 1 / self._G, opts_str)

    def current_equation(self, v, kind='t'):
        """Return expression for current through component given
        applied voltage."""

        return SuperpositionCurrent(SuperpositionVoltage(v).select(kind) / self._Z).select(kind)

    def voltage_equation(self, i, kind='t'):
        """Return expression for voltage across component given
        applied current."""

        return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Z).select(kind)


class NG(G):
    """Noiseless conductor"""

    is_noiseless = True


class L(OnePort):
    """Inductor

    Inductance Lval, initial current i0"""

    is_inductor = True

    def __init__(self, Lval='L', i0=None, **kwargs):

        self.kwargs = kwargs
        self.has_ic = i0 is not None
        if i0 is None:
            i0 = 0

        if self.has_ic:
            self.args = (Lval, i0)
        else:
            self.args = (Lval, )

        Lval = cexpr(Lval)
        i0 = cexpr(i0)
        self.L = Lval
        self.i0 = i0
        self._Z = impedance(s * Lval, causal=True)
        self._Voc = SuperpositionVoltage(LaplaceDomainExpression(-i0 * Lval))
        self.zeroic = self.i0 == 0

    def current_equation(self, v, kind='t'):
        """Return expression for current through component given
        applied voltage."""

        from .sym import tausym

        if kind in ('t', 'time', 'super'):
            u = tausym
            v = expr(v).subs(t, u)
            if self.has_ic:
                return SuperpositionCurrent(Integral(v, (u, 0, tsym)) / self.L).select(kind) + expr(self.i0)

            return SuperpositionCurrent(Integral(v, (u, -oo, tsym)) / self.L).select(kind)
        elif kind in ('s', 'laplace'):
            return SuperpositionCurrent((SuperpositionVoltage(v).select(kind) + self.L * self.i0) / self._Zkind(kind)).select(kind)
        return SuperpositionCurrent(SuperpositionVoltage(v).select(kind) / self._Zkind(kind)).select(kind)

    def voltage_equation(self, i, kind='t'):
        """Return expression for voltage across component given
        applied current."""

        if kind in ('t', 'time', 'super'):
            return SuperpositionVoltage(self.L * Derivative(i, t)).select(kind)
        elif kind in ('s', 'laplace'):
            return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Zkind(kind) - self.L * self.i0).select(kind)
        return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Zkind(kind)).select(kind)


class C(OnePort):
    """Capacitor

    Capacitance Cval, initial voltage v0"""

    is_capacitor = True

    def __init__(self, Cval='C', v0=None, **kwargs):

        self.kwargs = kwargs
        self.has_ic = v0 is not None
        if v0 is None:
            v0 = 0

        if self.has_ic:
            self.args = (Cval, v0)
        else:
            self.args = (Cval, )

        Cval = cexpr(Cval)
        v0 = cexpr(v0)
        self.C = Cval
        self.v0 = v0
        self._Z = impedance(1 / (s * Cval), causal=True)
        self._Voc = SuperpositionVoltage(LaplaceDomainExpression(v0 / s))
        self.zeroic = self.v0 == 0

    def current_equation(self, v, kind='t'):
        """Return expression for current through component given
        applied voltage."""

        if kind in ('t', 'time', 'super'):
            return SuperpositionCurrent(self.C * Derivative(v, t)).select(kind)
        elif kind in ('s', 'laplace'):
            return SuperpositionCurrent((SuperpositionVoltage(v).select(kind) - self.v0 / s) / self._Zkind(kind)).select(kind)

        return SuperpositionCurrent(SuperpositionVoltage(v).select(kind) / self._Zkind(kind)).select(kind)

    def voltage_equation(self, i, kind='t'):
        """Return expression for voltage across component given
        applied current."""

        from .sym import tausym

        if kind in ('t', 'time', 'super'):
            u = tausym
            i = expr(i).subs(t, u)
            if self.has_ic:
                return SuperpositionVoltage(Integral(i, (u, 0, tsym)) / self.C).select(kind) + expr(self.v0)

            return SuperpositionVoltage(Integral(i, (u, -oo, tsym)) / self.C).select(kind)
        elif kind in ('s', 'laplace'):
            return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Zkind(kind) + self.v0 / s).select(kind)

        return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Zkind(kind)).select(kind)


class CPE(OnePort):
    """Constant phase element

    This has an impedance 1 / (s**alpha * K).  When alpha == 0, the CPE is
    equivalent to a resistor of resistance 1 / K.  When alpha == 1, the CPE is
    equivalent to a capacitor of capacitance K.

    When alpha == 0.5 (default), the CPE is a Warburg element.

    The phase of the impedance is -pi * alpha / 2.

    Note, when alpha is non-integral, the impedance cannot be represented
    as a rational function and so there are no poles or zeros.  So
    don't be suprised if Lcapy throws an occasional wobbly."""

    def __init__(self, K, alpha=0.5, **kwargs):

        self.kwargs = kwargs
        self.args = (K, alpha)

        K = cexpr(K)
        alpha = cexpr(alpha)
        self.K = K
        self.alpha = alpha
        self._Z = impedance(1 / (s ** alpha * K), causal=True)


class Y(OnePort):
    """General admittance."""

    def __init__(self, Yval='Y', **kwargs):

        self.kwargs = kwargs
        self.args = (Yval, )
        Yval = admittance(Yval)
        self._Y = Yval

    def current_equation(self, v, kind='t'):
        """Return expression for current through component given
        applied voltage."""

        return SuperpositionCurrent(SuperpositionVoltage(v).select(kind) / self._Z).select(kind)

    def voltage_equation(self, i, kind='t'):
        """Return expression for voltage across component given
        applied current."""

        return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Z).select(kind)


class Z(OnePort):
    """General impedance."""

    def __init__(self, Zval='Z', **kwargs):

        self.kwargs = kwargs
        self.args = (Zval, )
        Zval = impedance(Zval)
        self._Z = Zval

    def current_equation(self, v, kind='t'):
        """Return expression for current through component given
        applied voltage."""

        return SuperpositionCurrent(SuperpositionVoltage(v).select(kind) / self._Z).select(kind)

    def voltage_equation(self, i, kind='t'):
        """Return expression for voltage across component given
        applied current."""

        return SuperpositionVoltage(SuperpositionCurrent(i).select(kind) * self._Z).select(kind)


class VoltageSourceBase(OnePort):

    is_voltage_source = True
    cpt_type = 'V'
    is_noisy = False

    def voltage_equation(self, i, kind='t'):
        """Return expression for voltage across component given
        applied current."""

        return SuperpositionVoltage(self.voc).select(kind)


class sV(VoltageSourceBase):
    """Arbitrary s-domain voltage source"""

    netkeyword = 's'

    def __init__(self, Vval, **kwargs):

        self.kwargs = kwargs
        self.args = (Vval, )
        Vval = LaplaceDomainExpression(Vval)
        self._Voc = SuperpositionVoltage(LaplaceDomainExpression(Vval))


class V(VoltageSourceBase):
    """Arbitrary voltage source"""

    def __init__(self, Vval='V', **kwargs):

        self.kwargs = kwargs
        self.args = (Vval, )
        self._Voc = SuperpositionVoltage(Vval)


class Vstep(VoltageSourceBase):
    """Step voltage source (s domain voltage of v / s)."""

    netkeyword = 'step'

    def __init__(self, v, **kwargs):

        self.kwargs = kwargs
        self.args = (v, )
        v = cexpr(v)
        self._Voc = SuperpositionVoltage(
            TimeDomainExpression(v) * Heaviside(t))
        self.v0 = v


class Vdc(VoltageSourceBase):
    """DC voltage source (note a DC voltage source of voltage V has
    an s domain voltage of V / s)."""

    netkeyword = 'dc'

    def __init__(self, Vval, **kwargs):

        self.kwargs = kwargs
        self.args = (Vval, )
        Vval = cexpr(Vval)
        self._Voc = SuperpositionVoltage(cexpr(Vval, dc=True))
        self.v0 = Vval

    @property
    def voc(self):
        return voltage(self.v0)


class Vac(VoltageSourceBase):
    """AC voltage source."""

    netkeyword = 'ac'

    def __init__(self, V, phi=None, omega=None, **kwargs):

        self.kwargs = kwargs
        if phi is None and omega is None:
            self.args = (V, )
        elif phi is not None and omega is None:
            self.args = (V, phi)
        elif phi is None and omega is not None:
            self.args = (V, 0, omega)
        else:
            self.args = (V, phi, omega)

        if phi is None:
            phi = 0

        if omega is None:
            omega = omega0sym
        omega = expr(omega)

        V = cexpr(V)
        phi = cexpr(phi)

        # Note, cos(-pi / 2) is not quite zero.

        self.omega = omega
        self.v0 = V
        self.phi = phi
        self._Voc = SuperpositionVoltage(phasor(self.v0 * exp(j * self.phi),
                                                omega=self.omega))

    @property
    def voc(self):
        return voltage(self.v0 * cos(self.omega * t + self.phi))


class Vnoise(VoltageSourceBase):
    """Noise voltage source."""

    netkeyword = 'noise'
    is_noisy = True

    def __init__(self, V, nid=None, **kwargs):

        self.kwargs = kwargs
        V1 = AngularFourierNoiseDomainVoltage(V, nid=nid)
        self.args = (V, V1.nid)
        self._Voc = SuperpositionVoltage(V1)


class v(VoltageSourceBase):
    """Arbitrary t-domain voltage source"""

    def __init__(self, vval, **kwargs):

        self.kwargs = kwargs
        self.args = (vval, )
        Vval = TimeDomainExpression(vval)
        self._Voc = SuperpositionVoltage(Vval)


class CurrentSourceBase(OnePort):

    is_current_source = True
    cpt_type = 'I'
    is_noisy = False

    @property
    def I(self):
        """Open-circuit current of a current source."""
        return self.Isc

    def current_equation(self, v, kind='t'):
        """Return expression for current through component given
        applied voltage."""

        return SuperpositionCurrent(self.isc).select(kind)


class sI(CurrentSourceBase):
    """Arbitrary s-domain current source"""

    netkeyword = 's'

    def __init__(self, Ival, **kwargs):

        if isinstance(Ival, str) and Ival == 'I':
            warn('Current I is being considered as the imaginary number')

        self.kwargs = kwargs
        self.args = (Ival, )
        Ival = LaplaceDomainExpression(Ival)
        self._Isc = SuperpositionCurrent(LaplaceDomainExpression(Ival))


class I(CurrentSourceBase):
    """Arbitrary current source"""

    def __init__(self, Ival='Is', **kwargs):

        if isinstance(Ival, str) and Ival == 'I':
            warn('Current I is being considered as the imaginary number')

        self.kwargs = kwargs
        self.args = (Ival, )
        self._Isc = SuperpositionCurrent(Ival)


class Istep(CurrentSourceBase):
    """Step current source (s domain current of i / s)."""

    netkeyword = 'step'

    def __init__(self, Ival, **kwargs):

        if isinstance(Ival, str) and Ival == 'I':
            warn('Current I is being considered as the imaginary number')

        self.kwargs = kwargs
        self.args = (Ival, )
        Ival = cexpr(Ival)
        self._Isc = SuperpositionCurrent(
            TimeDomainExpression(Ival) * Heaviside(t))
        self.i0 = Ival


class Idc(CurrentSourceBase):
    """DC current source (note a DC current source of current i has
    an s domain current of i / s)."""

    netkeyword = 'dc'

    def __init__(self, Ival, **kwargs):

        if isinstance(Ival, str) and Ival == 'I':
            warn('Current I is being considered as the imaginary number')

        self.kwargs = kwargs
        self.args = (Ival, )
        Ival = cexpr(Ival)
        self._Isc = SuperpositionCurrent(cexpr(Ival, dc=True))
        self.i0 = Ival

    @property
    def isc(self):
        return current(self.i0)


class Iac(CurrentSourceBase):
    """AC current source."""

    netkeyword = 'ac'

    def __init__(self, Ival, phi=0, omega=None, **kwargs):

        if isinstance(Ival, str) and Ival == 'I':
            warn('Current I is being considered as the imaginary number')

        self.kwargs = kwargs
        if phi is None and omega is None:
            self.args = (Ival, )
        elif phi is not None and omega is None:
            self.args = (Ival, phi)
        elif phi is None and omega is not None:
            self.args = (Ival, 0, omega)
        else:
            self.args = (Ival, phi, omega)

        if phi is None:
            phi = 0

        if omega is None:
            omega = omega0sym
        omega = cexpr(omega)

        Ival = cexpr(Ival)
        phi = cexpr(phi)

        self.omega = omega
        self.i0 = Ival
        self.phi = phi
        self._Isc = SuperpositionCurrent(phasor(self.i0 * exp(j * self.phi),
                                                omega=self.omega))

    @property
    def isc(self):
        return current(self.i0 * cos(self.omega * t + self.phi))


class Inoise(CurrentSourceBase):
    """Noise current source."""

    netkeyword = 'noise'
    is_noisy = True

    def __init__(self, Ival, nid=None, **kwargs):

        if isinstance(Ival, str) and Ival == 'I':
            warn('Current I is being considered as the imaginary number')

        self.kwargs = kwargs
        I1 = AngularFourierNoiseDomainCurrent(Ival, nid=nid)
        self._Isc = SuperpositionCurrent(I1)
        self.args = (Ival, I1.nid)


class i(CurrentSourceBase):
    """Arbitrary t-domain current source"""

    def __init__(self, Ival, **kwargs):

        if isinstance(Ival, str) and Ival == 'I':
            warn('Current I is being considered as the imaginary number')

        self.kwargs = kwargs
        self.args = (Ival, )
        Ival = TimeDomainExpression(Ival)
        self._Isc = SuperpositionCurrent(Ival)


class Xtal(OnePort):
    """Crystal

    This is modelled as a series R, L, C circuit in parallel
    with C0 (a Butterworth van Dyke model).  Note,
    harmonic resonances are not modelled.
    """

    def __init__(self, C0, R1, L1, C1, **kwargs):

        self.kwargs = kwargs
        self.C0 = cexpr(C0)
        self.R1 = cexpr(R1)
        self.L1 = cexpr(L1)
        self.C1 = cexpr(C1)

        self._Z = self.expand().impedance
        self.args = (C0, R1, L1, C1)

    def expand(self):

        return (R(self.R1) + L(self.L1) + C(self.C1)) | C(self.C0)

    def _net_make(self, netlist, n1=None, n2=None, dir='right'):

        # TODO: draw this with a symbol
        net = self.expand()
        return net._net_make(netlist, n1, n2, dir)


class FerriteBead(OnePort):
    """Ferrite bead (lossy inductor)

    This is modelled as a series resistor (Rs) connected
    to a parallel R, L, C network (Rp, Lp, Cp).
    """

    def __init__(self, Rs, Rp, Cp, Lp, **kwargs):

        self.kwargs = kwargs
        self.Rs = cexpr(Rs)
        self.Rp = cexpr(Rp)
        self.Cp = cexpr(Cp)
        self.Lp = cexpr(Lp)

        self._Z = self.expand().impedance
        self.args = (Rs, Rp, Cp, Lp)

    def expand(self):

        return R(self.Rs) + (R(self.Rp) + L(self.Lp) + C(self.Cp))

    def _net_make(self, netlist, n1=None, n2=None, dir='right'):

        # TODO: draw this with a symbol
        net = self.expand()
        return net._net_make(netlist, n1, n2, dir)


class LoadCircuit(Network):
    """Circuit comprised of a load oneport connected in parallel with a
    source oneport."""

    def __init__(self, source_OP, load_OP):

        self.source_OP = source_OP
        self.load_OP = load_OP

        self.vnet = source_OP | load_OP
        self.inet = source_OP + load_OP
        self.args = (source_OP, load_OP)

    @property
    def V(self):
        """Voltage across load."""
        return self.vnet.Voc

    @property
    def v(self):
        """Time-domain voltage across load."""
        return self.vnet.voc

    @property
    def I(self):
        """Current into load."""
        return self.inet.Isc

    @property
    def i(self):
        """Time-domain current into load."""
        return self.inet.isc

    def _net_make(self, netlist, n1=None, n2=None, dir='right'):

        # TODO: draw this better rather than as a oneport.
        return self.vnet._net_make(netlist, n1, n2, dir)


class ControlledSource(OnePort):
    """These components are controlled one-ports."""
    pass


class CCVS(ControlledSource):

    is_voltage_source = True

    def __init__(self, control, value, **kwargs):

        self.kwargs = kwargs
        self.args = (control, value)
        self._Voc = SuperpositionVoltage(0)
        self._Z = impedance(0)


class CCCS(ControlledSource):

    is_current_source = True

    def __init__(self, control, value, **kwargs):

        self.kwargs = kwargs
        self.args = (control, value)
        self._Isc = SuperpositionCurrent(0)
        self._Y = admittance(0)


class VCVS(ControlledSource):

    is_voltage_source = True

    def __init__(self, value, Ac=0, **kwargs):

        self.kwargs = kwargs
        self.args = (value, Ac)
        self._Voc = SuperpositionVoltage(0)
        self._Z = impedance(0)


class VCCS(ControlledSource):

    is_current_source = True

    def __init__(self, value, **kwargs):

        self.kwargs = kwargs
        self.args = (value, )
        self._Isc = SuperpositionCurrent(0)
        self._Y = admittance(0)


class Dummy(OnePort):

    def __init__(self, *args, **kwargs):

        self.kwargs = kwargs
        self.args = args


class K(Dummy):
    """Coupling coefficient"""

    def __init__(self, L1, L2, K, **kwargs):

        if K is ksym or (isinstance(K, Expr) and K.var is ksym):
            warn("""
Coupling coefficient %s is the discrete Fourier domain variable.
You can override it using %s = symbol('%s', force=True).""" % (K, K, K))

        self.kwargs = kwargs
        self.args = (L1, L2, K)
        self.K = cexpr(K)


class W(Dummy):
    """Wire (short)"""

    def __init__(self, **kwargs):

        self.kwargs = kwargs
        self.args = ()
        self._Z = impedance(0)


class O(Dummy):
    """Open circuit"""

    def __init__(self, **kwargs):

        self.kwargs = kwargs
        self.args = ()
        self._Y = admittance(0)


class P(O):
    """Port (open circuit)"""
    pass


class Mass(C):
    """Mass

    Mass mval, initial velocity v0"""
    pass


class Spring(L):
    """Spring

    Spring constant kval, initial force f0"""

    def __init__(self, kval='k', f0=None, **kwargs):

        Lval = 1 / cexpr(kval)

        super().__init__(Lval, f0, **kwargs)


class Damper(G):
    """Damper

    Friction coeff val"""
    pass


def series(*args):
    """Create a series combination of a number of components.
    Args that are None are ignored.  If there is only one
    non-None component, return that component."""

    args = [net for net in args if net is not None]
    if args == []:
        return None
    if len(args) == 1:
        return args[0]
    return Ser(*args)


def parallel(*args):
    """Create a parallel combination of a number of components.
    Args that are None are ignored.  If there is only one
    non-None component, return that component."""

    args = [net for net in args if net is not None]
    if args == []:
        return None
    if len(args) == 1:
        return args[0]
    return Par(*args)


def ladder(*args, **kwargs):
    """Create a ladder oneport network with alternating series and shunt components.
    If an arg is None, the component is ignored.

    ladder(R(1), C(2), R(3)) is equivalent to R(1) + (C(1) | R(3))

    ladder(None, R(1), C(2), R(3)) is equivalent to R(1) | (C(1) + R(3))
    """

    start_series = kwargs.pop('start_series', True)

    if len(args) == 0:
        return None
    elif len(args) == 1:
        return args[0]
    elif start_series:
        return series(args[0], ladder(*args[1:], start_series=not start_series))
    else:
        return parallel(args[0], ladder(*args[1:], start_series=not start_series))


# Imports at end to circumvent circular dependencies
from .twoport import Ladder, LSection, TSection  # nopep8
from .superpositioncurrent import SuperpositionCurrent  # nopep8
from .superpositionvoltage import SuperpositionVoltage  # nopep8
from .noiseomegaexpr import AngularFourierNoiseDomainCurrent, AngularFourierNoiseDomainVoltage  # nopep8
from .phasor import phasor  # nopep8
from .texpr import TimeDomainExpression  # nopep8
from .sexpr import LaplaceDomainExpression  # nopep8
from .cexpr import cexpr  # nopep8
from .expr import Expr, expr  # nopep8



from lcapy import (symbol, nexpr, cos, sin, exp, j, sinc,
                   rect, sincu, n, dt, delta, sign, H, pi, f)

alpha = symbol('alpha')
m = symbol('m', integer=True)
f0 = symbol('f0')
w0 = 2 * pi * f0
No = symbol('N_o', odd=True)
Ne = symbol('N_e', even=True)
K = symbol('K')

sigs = [nexpr('x(n)'), nexpr('x(a * n)'), nexpr('x(n - m)'),
        cos(w0 * n * dt), sin(w0 * n * dt), exp(j * w0 * n * dt),
        nexpr(1), delta(n), delta(n - m),
        H(n), n * H(n), sign(n),
        alpha**-n * H(n), rect(n), rect(n / No), rect(n / Ne),
        sinc(n), sinc(K * n), sinc(K * n)**2, sincu(K * n)]

for sig in sigs:
    print(':math:`%s \\longleftrightarrow %s`\n' %
          (sig.latex(), sig(f).latex()))



from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_loop1(self):

        a = Circuit("""
        V 1 0 {v(t)}
        R 1 0""")

        la = a.loop_analysis()

        la_eqs = la.mesh_equations()
        loops = la.loops_by_cpt_name()

        key = list(la_eqs.keys())[0]
        eq = la_eqs[key]

        self.assertEqual(len(loops), 1, 'num loops')

        loop = loops[0]

        self.assertEqual(eq.lhs, voltage('v(t) - R * i_1(t)'),
                         'mesh_equations()[1].lhs')

        self.assertEqual(eq.rhs, voltage(0), 'mesh_equations()[1].rhs')

    def test_loop2(self):

        a = Circuit("""
        V1 1 0 {v1(t)}
        V2 1 2 {v2(t)}
        R 2 0""")

        la = a.loop_analysis()

        la_eqs = la.mesh_equations()
        loops = la.loops_by_cpt_name()

        key = list(la_eqs.keys())[0]
        eq = la_eqs[key]

        self.assertEqual(len(loops), 1, 'num loops')

        loop = loops[0]

        self.assertEqual(eq.lhs,
                         voltage('v1(t) - v2(t) - R * i_1(t)'),
                         'mesh_equations()[1].lhs')

        self.assertEqual(eq.rhs, voltage(0), 'mesh_equations()[1].rhs')

    def test_loop3(self):

        a = Circuit("""
        V 1 0 {v(t)}
        R1 1 2
        R2 2 3
        R3 3 0_3
        W 0 0_3
        W 3 3_a
        R4 3_a 0_4
        W 0_3 0_4""")

        la = a.loop_analysis()

        la_eqs = la.mesh_equations()
        loops = la.loops_by_cpt_name()

        key = list(la_eqs.keys())[0]
        eq = la_eqs[key]
        loop = loops[0]

        if loop == ['R2', 'C']:
            self.assertEqual(abs(eq.lhs), voltage(
                'abs(R3*(i_2(t) - i_1(t)) + R4*(i_2(t) - i_1(t)) + v(t))'), 'mesh_equations()[1].lhs')
        elif loop == ['V', 'R1', 'R2', 'R3']:
            self.assertEqual(abs(eq.lhs), voltage(
                'abs(-R1*i_1(t) - R2*i_1(t) + R3*(-i_1(t) + i_2(t)) + v(t))'), 'mesh_equations()[1].lhs')
        elif loop == ['R3', 'R4']:
            self.assertEqual(abs(eq.lhs), voltage(
                'abs(R3*(i_1(t) - i_2(t)) + R4*(i_1(t) - i_2(t)))'), 'mesh_equations()[1].lhs')
        else:
            raise ValueError('Unhandled loop %s' % loop)

        # Use abs since sometimes the current flows in the other direction
        # depending how the loop was chosen.

        self.assertEqual(eq.rhs, voltage(0), 'mesh_equations()[1].rhs')



"""This module provides transfer function support.

Copyright 2020--2023 Michael Hayes, UCECE

"""
from .expr import expr
from .units import u as uu


def transfer(numer, denom=1, causal=True, **assumptions):
    """Create transfer function assuming zero initial conditions
    from the ratio `numer / denom`."""

    numer = expr(numer, causal=causal, **assumptions)
    denom = expr(denom, causal=causal, **assumptions)

    numer = numer.zero_initial_conditions()
    denom = denom.zero_initial_conditions()
    expr1 = numer / denom

    if expr1.is_admittance:
        expr1.units = 1 / uu.ohms
        return expr1

    try:
        expr1 = expr1.as_transfer()
    except:
        raise ValueError('Cannot represent %s(%s) as transfer function' % (
            expr1.__class__.__name__, expr1))

    return expr1
